# Copyright (c) 2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#-------------------------------------------------------------------------------
#
#   usage: python -m test.test_package -v
#          python -m test.test_package PackageTest.<method> -v
#

import unittest
import os
import sys
import shutil
import glob
import tempfile
import subprocess

from afsutil.package import RpmBuilder
from afsutil.package import MockRpmBuilder

#REPO = os.path.expanduser("~/src/openafs")
REPO = "https://github.com/openafs/openafs.git"
TAG = "openafs-stable-1_6_20"
SRCDIR = tempfile.mkdtemp()

# Test mode
SKIP_BUILDS = True
SKIP_TEARDOWN = False

def readfile(path):
    with open(path, 'r') as f:
        return f.read()

def writefile(path, contents):
    with open(path, 'w') as f:
        f.write(contents)

def setup_srcdir():
    """Clone the source tree."""
    script = """
        mkdir -p {srcdir}
        cd {srcdir}
        git init
        git remote add origin {repo}
        git fetch origin
        git checkout {tag}""".\
        format(srcdir=SRCDIR, repo=REPO, tag=TAG)
    rc = os.system("/bin/bash -x -e -c '{script}'".format(script=script))
    if rc != 0:
        shutil.rmtree(SRCDIR)
        raise Exception("Unable to create srcdir.")

def kdevel():
    """Get list of installed kernel-devel packages."""
    return subprocess.check_output([
        "rpm", "-q", "-a", "kernel-devel"]).splitlines()

class PackageTest(unittest.TestCase):

    srcdir = SRCDIR
    topdir = None
    dstdir = None

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def setUpClass(cls):
        setup_srcdir()
        if not SKIP_BUILDS:
            sys.stdout.write("Note: These test will take a long, long time.\n")
            sys.stdout.write("      Set SKIP_BUILDS to True to skip long running tests.\n")

    @classmethod
    def tearDownClass(cls):
        if SKIP_TEARDOWN:
            sys.stdout.write("Warning: Teardown skipped; remove test dirs manually")
            return
        if os.path.isdir(SRCDIR):
             shutil.rmtree(SRCDIR)
        if cls.topdir is not None and os.path.isdir(cls.topdir):
            shutil.rmtree(cls.topdir)
        if cls.dstdir is not None and os.path.isdir(cls.dstdir):
            shutil.rmtree(cls.dstdir)

    def setUp(self):
        self.topdir = tempfile.mkdtemp()
        self.dstdir = tempfile.mkdtemp()

    def tearDown(self):
        if self.topdir is not None and os.path.isdir(self.topdir):
            shutil.rmtree(self.topdir)
        if self.dstdir is not None and os.path.isdir(self.dstdir):
            shutil.rmtree(self.dstdir)

    def assertExists(self, path):
        self.assertIsNotNone(path)
        self.assertTrue(os.path.exists(path),
            msg="path {0} does not exist".format(path))

    def assertNotExists(self, path):
        self.assertIsNotNone(path)
        self.assertFalse(os.path.exists(path),
            msg="path {0} exists".format(path))

    def assertGlobExists(self, g):
        files = glob.glob(g)
        self.assertTrue(len(files) > 0, msg="No files found for glob {0}".format(g))

    def test_01_get_package_version(self):
        testcases = [
            ('1.6.0', '1.6.0', '1'),
            ('1.6.0pre1', '1.6.0', '0.pre1'),
            ('1.8.0dev', '1.8.0', '0.dev'),
            ('1.6.20-32-gabcdef', '1.6.20', '32.gabcdef'),
            ('1.6.18.3-foo3', '1.6.18.3', '1.2.3.foo'),
        ]
        b = RpmBuilder()
        for tc in testcases:
            (v, r) = b.get_package_version(version=tc[0])
            self.assertEquals(v, tc[1])
            self.assertEquals(r, tc[2])

    def test_02_get_version(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        version = b.get_version()
        self.assertIsNotNone(version)
        self.assertRegexpMatches(version, r'^[1-9]\.[0-9]+\.[0-9]+.*')

    def test_03_get_version__from_dot_version(self):
        dot_version = os.path.join(self.srcdir, ".version")
        writefile(dot_version, "1.2.3")
        try:
            b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
            version = b.get_version()
            self.assertIsNotNone(version)
            self.assertEquals(version, "1.2.3")
        finally:
            os.remove(dot_version)

    def test_04_generate_spec(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        infile = "{srcdir}/src/packaging/RedHat/openafs.spec.in".format(srcdir=self.srcdir)
        spec = b.generate_spec(infile)
        self.assertEquals(spec, "{topdir}/SPECS/openafs.spec".format(topdir=self.topdir))
        self.assertExists("{topdir}/SPECS/openafs.spec".format(topdir=self.topdir))

    def test_05_copy_spec(self):
        spec = "/tmp/openafs.spec"
        writefile(spec, "example")
        try:
            b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
            b.copy_spec(spec)
            self.assertExists("{topdir}/SPECS/openafs.spec".format(topdir=self.topdir))
        finally:
            os.remove(spec)

    def test_06_prepare_spec(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        b.prepare_spec()
        self.assertExists("{topdir}/SPECS/openafs.spec".format(topdir=self.topdir))

    def test_07_prepare_spec__external_spec(self):
        spec = "/tmp/openafs.spec"
        writefile(spec, "example")
        try:
            b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir, spec=spec)
            b.prepare_spec()
            self.assertExists("{topdir}/SPECS/openafs.spec".format(topdir=self.topdir))
            contents = readfile("{topdir}/SPECS/openafs.spec".format(topdir=self.topdir))
            self.assertEquals(contents, "example")
        finally:
            os.remove(spec)

    def test_08_list_sources(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        s = b.list_sources()
        self.assertIsNotNone(s)

    def test_09_lookup_source(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        s = b.lookup_source('CellServDB')
        self.assertIsNotNone(s)

    def test_10_download_csdb(self):
        url = "https://www.central.org/dl/cellservdb/CellServDB.2016-01-01"
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        b.download_csdb(url)
        self.assertExists("{topdir}/SOURCES/CellServDB.2016-01-01".format(topdir=self.topdir))

    def test_11_copy_csdb(self):
        csdb = "/tmp/CellServDB"
        writefile(csdb, "example")
        try:
            b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
            b.copy_csdb(csdb)
            self.assertExists("{topdir}/SOURCES/CellServDB".format(topdir=self.topdir))
        finally:
            os.remove(csdb)

    def test_12_prepare_sources_csdb(self):
        csdb = "/tmp/CellServDB"
        writefile(csdb, "example")
        try:
            b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir, csdb=csdb)
            b.prepare_sources_csdb()
            self.assertExists("{topdir}/SOURCES/CellServDB".format(topdir=self.topdir))
        finally:
            os.remove(csdb)

    def test_13_prepare_sources_relnotes(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        b.prepare_sources_relnotes()
        version = b.get_version()
        self.assertExists("{topdir}/SOURCES/RELNOTES-{version}".format(topdir=self.topdir,version=version))

    def test_14_prepare_sources_changelog(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        b.prepare_sources_changelog()
        self.assertExists("{topdir}/SOURCES/ChangeLog".format(topdir=self.topdir))

    def test_15_prepare_sources_scripts(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        b.prepare_sources_scripts()
        self.assertGlobExists("{topdir}/SOURCES/*".format(topdir=self.topdir))

    def test_16_prepare_sources_tarballs(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir)
        b.prepare_sources_tarballs()
        self.assertGlobExists("{topdir}/SOURCES/*.tar.*".format(topdir=self.topdir))

    def test_17_prepare_sources(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir, quiet=True)
        b.prepare_sources()
        self.assertGlobExists("{topdir}/SOURCES/*".format(topdir=self.topdir))

    def test_18_build_srpm(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir, quiet=True)
        b.build_srpm()
        self.assertGlobExists("{topdir}/SRPMS/*".format(topdir=self.topdir))

    @unittest.skipIf(SKIP_BUILDS, "quick mode")
    def test_19_build_userspace(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir, quiet=True)
        b.build_userspace()
        self.assertGlobExists("{topdir}/RPMS/*".format(topdir=self.topdir))

    @unittest.skipIf(SKIP_BUILDS, "quick mode")
    @unittest.skipUnless(len(kdevel()) > 0, "requires kernel-devel packages")
    def test_20_build_kmods(self):
        b = RpmBuilder(srcdir=self.srcdir, topdir=self.topdir, quiet=True)
        b.build_kmods()
        self.assertGlobExists("{topdir}/RPMS/*".format(topdir=self.topdir))

    @unittest.skipIf(SKIP_BUILDS, "quick mode")
    def test_21_build_userspace_with_mock(self):
        b = MockRpmBuilder(
                'epel-7-x86_64',
                srcdir=self.srcdir,
                dstdir=self.dstdir,
                autoclean=True, quiet=True)
        b.build_userspace()
        self.assertGlobExists("{dstdir}/openafs*.rpm".format(dstdir=self.dstdir))

    @unittest.skipIf(SKIP_BUILDS, "quick mode")
    def test_21_build_kmods_with_mock(self):
        b = MockRpmBuilder(
                'epel-7-x86_64',
                srcdir=self.srcdir,
                dstdir=self.dstdir,
                autoclean=True, quiet=True)
        b.build_kmods()
        self.assertGlobExists("{dstdir}/kmod-openafs*.rpm".format(dstdir=self.dstdir))

if __name__ == "__main__":
     unittest.main()




# This is a generated file. Do not edit.

resources = {
    "openafs-client-linux.init":\
"""\
#! /bin/bash
# Copyright 2000, International Business Machines Corporation and others.
# All Rights Reserved.
#
# This software has been released under the terms of the IBM Public
# License.  For details, see the LICENSE file in the top-level source
# directory or online at http://www.openafs.org/dl/license10.html

# AFS	Start and stop AFS components
#
#
# chkconfig: 35 60 20
# description:  AFS is a distributed file system which provides location
#		transparency, caching and secure authentication.
#		Additional configuration can be done in the /etc/sysconfig/afs
#		file. Read the documentation in that file for more information.
#
# The following section is used at least by SuSE insserv(8)
### BEGIN INIT INFO
# Provides: afs
# Required-Start: $network
# Required-Stop: $network
# Default-Start: 3 5
# Default-Stop: 0 1 2 6
# Description:  AFS is a distributed file system which provides location
#              transparency, caching and secure authentication.
#              Additional configuration can be done in the /etc/sysconfig/afs
#              file. Read the documentation in that file for more information.
### END INIT INFO
#
# Note that AFS does not use a pid file in /var/run. It is turned off by
# unmounting /afs.


# check for special bootup functions
if [ -f /etc/rc.d/init.d/functions ] ; then
	. /etc/rc.d/init.d/functions
	afs_rh=1
else
	# special (RedHat) functions not available...
	# so I provide neccessary replacements (e.g. for SuSE)

	function echo_failure () { echo -n " - failed." ; }

	function echo_success () { echo -n " - successful." ; }

	# these are hopefully all functions I use...
fi

runcmd() {
   echo -n "$1 "
   shift
   if [ "$BOOTUP" = "color" ]; then
      $* && echo_success || echo_failure
   else
      $*
   fi
   echo
}

KERNEL_VERSION=`uname -r`

# Gather up options and post startup script name, if present
SYSCNF=/etc/sysconfig/openafs-client
CACHESIZE="AUTOMATIC"
AFSD_OPTIONS=""
CACHEINFO=/usr/vice/etc/cacheinfo
CACHE=/usr/vice/cache
AFS=/afs
AFSD=/usr/vice/etc/afsd
KILLAFS=/usr/vice/etc/killafs
if [ -f $SYSCNF ] ; then
	. $SYSCNF
fi

on_network() {
    if hash ip 2>/dev/null >/dev/null; then
        IP="ip -4 addr"
    elif hash ifconfg 2>/dev/null >/dev/null; then
        IP="ifconfig -a"
    else
        echo afs: Unable to check interfaces. 1>&2
        return 1
    fi
    ADDRS=`LANG=C ${IP} | grep 'inet ' | grep -v 127.0.0.1 | wc -l`
    if [ "$ADDRS" = "" ]; then
       echo afs: No interfaces with IP address 1>&2
       return 1
    elif [ $ADDRS = 0 ]; then
       echo afs: No interfaces with IP address 1>&2
       return 1
    fi
    return 0
}

# If choose_client can't correctly determine which client to use, set
# LIBAFS manually.
choose_client() {

	# Use the second field of the uname -v output instead of just
	# doing a match on the whole thing to protect against matching
	# a timezone named SMP -- I don't know of one, but let's be
	# paranoid.

	set X `uname -v`; shift
	case $2 in
	SMP) MP=.mp ;;	# MP system
	*)   MP= ;;	# SP system
	esac

	# For now, just use uname -r to get the module version.
	case $KERNEL_VERSION in
	  [1-2].[0-5].*)
		LIBAFS=libafs-$KERNEL_VERSION$MP.o
		;;
	  *)
		LIBAFS=libafs-$KERNEL_VERSION$MP.ko
		;;
	esac
}

#
# Find prefix symbol to use with insmod.  We find the unregister_filesystem
# string from /proc/ksyms since we know it's there.  If /proc/ksyms does not
# exist, we print that info to the console and use the uname -v output to
# decide on a prefix.
# unregister_filesystem_Rsmp_b240cad8 is a typcial SMP version string from
# a kernel built from ftp.kernel.org
#
case $KERNEL_VERSION in
  [1-2].[0-5].*)
	KSYMS_FILE=/proc/ksyms
	;;
  *)
	KSYMS_FILE=/proc/kallsyms
	;;
esac
SEARCH_STR="unregister_filesystem"
DEFAULT_SMP_PREFIX="smp_" # Redhat kernels need "smp" instead
PREFIX="" # none needed for UP with <= 1Gig memory

set_prefix()
{
	h='[0-9a-fA-F]'
	h8="$h$h$h$h$h$h$h$h"
	prefix_set=0

	set X `egrep "\<$SEARCH_STR" $KSYMS_FILE 2> /dev/null`; shift

	case $KERNEL_VERSION in
	  [1-2].[0-5].*)
		str=$2
		;;
	  *)
		str=$3
		;;
	esac
	case $str in
	${SEARCH_STR}_R$h8)
		# No prefix required
		;;
	$SEARCH_STR)
		# No versioning in kernel symbols
		;;
	${SEARCH_STR}_R*$h8)
		suffix=${str#${SEARCH_STR}_R}
		PREFIX=${suffix%$h8}
		;;
	*)
		case $str in
		'')
			echo afsd: Cannot find \"$SEARCH_STR\" in file $KSYMS_FILE
			;;
		*)
			echo afsd: Malformed kernel version symbol \"$str\"
			;;
		esac

		echo Guessing prefix from output of uname -v
		set X `uname -v`; shift
		case $2 in
		SMP)
			PREFIX=$DEFAULT_SMP_PREFIX
			;;
		esac
		;;
	esac
}

MODLOADDIR=${MODLOADDIR:-/usr/vice/etc/modload}
# load_client loads the AFS client module if it's not already loaded.
load_client() {

	# Load the packaging style kmod if present, otherwise fallback
	# to the legacy style.
	if [ -f /lib/modules/`uname -r`/extra/openafs/openafs.ko ] ; then
		modprobe openafs
		return
	fi

	# If LIBAFS is set, use it.
	if [ -z "$LIBAFS" ] ; then
		# Try to determine the right client.
		choose_client
	fi

	if [ ! -f "$MODLOADDIR/$LIBAFS" ] ; then
		echo AFS module $MODLOADDIR/$LIBAFS does not exist. Not starting AFS.
		exit 1
	fi

	# We need exportfs in order to access the cache files. Load it, but
	# ignore errors as on some machines it may be built in to the kernel.
	/sbin/modprobe exportfs >/dev/null 2>&1

	if [ -f $KSYMS_FILE ]; then
		# use the prefix command if required
		case $KERNEL_VERSION in
		  [1-2].[0-5].*)
			set_prefix
			/sbin/insmod ${PREFIX:+-P $PREFIX} -f -m $MODLOADDIR/$LIBAFS > $MODLOADDIR/libafs.map 2>&1
			;;
		  *)
			/sbin/insmod $MODLOADDIR/$LIBAFS > $MODLOADDIR/libafs.map 2>&1
			;;
		esac
	else
		/sbin/insmod $MODLOADDIR/$LIBAFS > $MODLOADDIR/libafs.map 2>&1
	fi
}

generate_cacheinfo() {
    if [ "$CACHESIZE" = "AUTOMATIC" ]; then
	LINE=`df -k $CACHE | tail -1`
	PART=`echo $LINE | awk '{ if ( ($NF != "/usr")  && ($NF != "/") ) print $NF; else print "NONE";}'`
	if [ "$PART" = "NONE" ]; then
	    echo "$CACHE or /usr/vice is not a separate partition"
	    echo "you have to change the cachesize in $SYSCNF by hand"
	    echo "AFS will be started with a VERY small cache of 8Mb."
	    CACHESIZE=8000
	else
	    # Check to see if df has pretty-printed for long dev (i.e. LVM)
            FCHAR=`echo $LINE | cut -c 1`
            if [ "$FCHAR" = "/" ]; then
                PARTSIZE=`echo $LINE | awk '{print $2}'`
            else
                PARTSIZE=`echo $LINE | awk '{print $1}'`
	    fi
	    CACHESIZE=`echo $PARTSIZE | awk '{printf "%d",int(($1*.8)/1000)*1000}'`
	fi
    fi
    if [ "x$CACHESIZE" != "x" ]; then
	echo $AFS:$CACHE:$CACHESIZE >$CACHEINFO
	chmod 0644 $CACHEINFO
    else
	CACHESIZE=`awk -F: '{print $3}' < $CACHEINFO`
    fi
}

generate_csdb() {
    if [ -f /usr/vice/etc/CellServDB.local -a -f /usr/vice/etc/CellServDB.dist ]; then
        if [ -h /usr/vice/etc/CellServDB ]; then
            rm -f /usr/vice/etc/CellServDB
        fi
        cat /usr/vice/etc/CellServDB.local /usr/vice/etc/CellServDB.dist >/usr/vice/etc/CellServDB
        chmod 644 /usr/vice/etc/CellServDB
    fi
}

case "$1" in
  start)
    if [ ! "$afs_rh" -o ! -f /var/lock/subsys/openafs-client ]; then
        if [ `echo "$AFSD_OPTIONS" | grep -c dynroot` = 0 ]; then
            on_network || exit 1
        fi
	# Load kernel extensions
	if  load_client  ; then :
	else
		echo Failed to load AFS client, not starting AFS services.
		exit 1
	fi

	echo "Starting AFS services..... "
	generate_cacheinfo
	generate_csdb
	${AFSD} ${AFSD_OPTIONS}
	test "$afs_rh" && touch /var/lock/subsys/openafs-client
	$AFS_POST_INIT
    fi
	;;

  stop)
    if [ ! "$afs_rh" -o -f /var/lock/subsys/openafs-client ]; then
	# Stop AFS
	echo "Stopping AFS services..... "

	if [ -x $KILLAFS ] ; then
		runcmd "Sending all processes using /afs the TERM signal ..." $KILLAFS TERM
		runcmd "Sending all processes using /afs the KILL signal ..." $KILLAFS KILL
	fi
	umount /afs

	# Unload the packaging style kmod if present, otherwise fallback
	# to the legacy style.
	if [ -f /lib/modules/`uname -r`/extra/openafs/openafs.ko ] ; then
		modprobe -r openafs
	else
		LIBAFS=`/sbin/lsmod | fgrep 'libafs'`
		if [ -n "$LIBAFS" ] ; then
			LIBAFS=`echo $LIBAFS | awk 'BEGIN { FS = " " } { print $1 }'`
			/sbin/rmmod $LIBAFS
		fi
	fi

	rm -f /var/lock/subsys/openafs-client
    fi
	;;

  restart)
	# Restart AFS
	$0 stop
	$0 start
	;;

  *)
	echo Usage: 'afs <start|stop|restart>'

esac

exit 0

action fool the Red Hat initscripts
""",


    "openafs-client-solaris-5.10.init":\
"""\
#!/bin/sh
#
# afs.rc: rc script for AFS on Solaris 10 platforms
#
# Install this script as /etc/init.d/afs.rc
# then make links like this:
# ln -s ../init.d/afs.rc /etc/rc0.d/K66afs
# ln -s ../init.d/afs.rc /etc/rc2.d/S70afs
#
CONFIG=/usr/vice/etc/config
AFSDOPT=$CONFIG/afsd.options
PACKAGE=$CONFIG/package.options

# EXTRAOPTS can be used to enable/disable AFSDB support (-afsdb)
# and Dynroot (dynamically-generated /afs) support (-dynroot).
EXTRAOPTS="-afsdb"

LARGE="-stat 2800 -dcache 2400 -daemons 5 -volumes 128"
MEDIUM="-stat 2000 -dcache 800 -daemons 3 -volumes 70"
SMALL="-stat 300 -dcache 100 -daemons 2 -volumes 50"

if [ -f $AFSDOPT ]; then
    OPTIONS=`cat $AFSDOPT`
else
    OPTIONS="$MEDIUM $EXTRAOPTS"
fi

# Need the commands ps, awk, kill, sleep
PATH=${PATH}${PATH:+:}/sbin:/bin:/usr/bin

killproc() {            # kill the named process(es)
      awkfield2='$2'
        pid=`ps -ef | awk "/$1/ && ! /awk/ {print $awkfield2}"`
        [ "$pid" != "" ] && kill -KILL $pid
}

generate_csdb() {
    if [ -f /usr/vice/etc/CellServDB.local -a -f /usr/vice/etc/CellServDB.dist ]; then
        if [ -h /usr/vice/etc/CellServDB ]; then
            rm -f /usr/vice/etc/CellServDB
        fi
        cat /usr/vice/etc/CellServDB.local /usr/vice/etc/CellServDB.dist >/usr/vice/etc/CellServDB
        chmod 644 /usr/vice/etc/CellServDB
    fi
}

case $1 in
'start')

#
# Make sure afs exists in /etc/name_to_sysnum
#
if grep -s "afs" /etc/name_to_sysnum > /dev/null; then
    echo "Entry for afs already exists in /etc/name_to_sysnum"
else
    echo "Creating entry for afs in /etc/name_to_sysnum"
    cp /etc/name_to_sysnum /etc/name_to_sysnum.orig
    sed '/nfs/i\
afs			65' /etc/name_to_sysnum > /tmp/name_to_sysnum
    mv /tmp/name_to_sysnum /etc/name_to_sysnum
    echo "Rebooting now for new /etc/name_to_sysnum to take effect"
    reboot
fi

## Check to see that /bin/isalist exists and is executable
if [ ! -x /bin/isalist ] ;then
      echo "/bin/isalist not executable"
      exit 1;
fi

## Determine if we are running the 64 bit OS
## If sparcv9 then the location of the afs and nfs extensions differ

case `/bin/isalist` in
    *amd64* )
              nfssrv=/kernel/misc/amd64/nfssrv
              afs=/kernel/fs/amd64/afs ;;
    *sparcv9* )
              nfssrv=/kernel/misc/sparcv9/nfssrv
              afs=/kernel/fs/sparcv9/afs ;;
          * )
              nfssrv=/kernel/misc/nfssrv
              afs=/kernel/fs/afs ;;
esac


#
# Load kernel extensions
#
# nfssrv has to be loaded first


if [ -f $nfssrv ]; then
      echo "Loading NFS server kernel extensions"
      modload $nfssrv
else
      echo "$nfssrv does not exist. Skipping AFS startup."
      exit 1
fi

## Load AFS kernel extensions

if [ -f $afs ]; then
      echo "Loading AFS kernel extensions"
      modload $afs
else
      echo "$afs does not exist. Skipping AFS startup."
      exit 1
fi

#
# Check that all of the client configuration files exist
#
generate_csdb
for file in /usr/vice/etc/afsd /usr/vice/etc/cacheinfo \
          /usr/vice/etc/ThisCell /usr/vice/etc/CellServDB
do
      if [ ! -f ${file} ]; then
              echo "${file} does not exist. Not starting AFS client."
              exit 1
      fi
done

#
# Check that the root directory for AFS (/afs)
# and the cache directory (/usr/vice/cache) both exist
#
for dir in `awk -F: '{print $1, $2}' /usr/vice/etc/cacheinfo`
do
      if [ ! -d ${dir} ]; then
              echo "${dir} does not exist. Not starting AFS client."
              exit 2
      fi
done

echo "Starting afsd"
/usr/vice/etc/afsd $OPTIONS

echo ;;

'stop')

afsroot=`awk -F: '{print $1}' /usr/vice/etc/cacheinfo` || exit 1
if /usr/sbin/mount | grep -q "$afsroot"; then
    echo "Unmounting $afsroot"
    /usr/sbin/umount $afsroot || exit 1

    echo "Stopping afsd"
    /usr/vice/etc/afsd -shutdown || exit 1

    afsid=`modinfo | awk '/afs filesystem/ {print $1}'` || exit 1
    echo "Unloading module id $afsid"
    modunload -i "$afsid"  || exit 1
fi
echo ;;

*)    echo "Invalid option supplied to $0"
      exit 1;;
esac
""",


    "openafs-client-solaris-5.11.init":\
"""\
#!/bin/sh
#
# afs.rc: rc script for AFS on Solaris 11 or OpenSolaris-based platforms
#
# Install this script as /etc/init.d/afs.rc
# then make links like this:
# ln -s ../init.d/afs.rc /etc/rc0.d/K66afs
# ln -s ../init.d/afs.rc /etc/rc2.d/S70afs
#
CONFIG=/usr/vice/etc/config
AFSDOPT=$CONFIG/afsd.options
PACKAGE=$CONFIG/package.options

# EXTRAOPTS can be used to enable/disable AFSDB support (-afsdb)
# and Dynroot (dynamically-generated /afs) support (-dynroot).
EXTRAOPTS="-afsdb"

LARGE="-stat 2800 -dcache 2400 -daemons 5 -volumes 128"
MEDIUM="-stat 2000 -dcache 800 -daemons 3 -volumes 70"
SMALL="-stat 300 -dcache 100 -daemons 2 -volumes 50"

if [ -f $AFSDOPT ]; then
    OPTIONS=`cat $AFSDOPT`
else
    OPTIONS="$MEDIUM $EXTRAOPTS"
fi

# Need the commands ps, awk, kill, sleep
PATH=${PATH}${PATH:+:}/sbin:/bin:/usr/bin

killproc() {            # kill the named process(es)
      awkfield2='$2'
        pid=`ps -ef | awk "/$1/ && ! /awk/ {print $awkfield2}"`
        [ "$pid" != "" ] && kill -KILL $pid
}

generate_csdb() {
    if [ -f /usr/vice/etc/CellServDB.local -a -f /usr/vice/etc/CellServDB.dist ]; then
        if [ -h /usr/vice/etc/CellServDB ]; then
            rm -f /usr/vice/etc/CellServDB
        fi
        cat /usr/vice/etc/CellServDB.local /usr/vice/etc/CellServDB.dist >/usr/vice/etc/CellServDB
        chmod 644 /usr/vice/etc/CellServDB
    fi
}


case $1 in
'start')

## Check to see that /bin/isalist exists and is executable
if [ ! -x /bin/isalist ] ;then
      echo "/bin/isalist not executable"
      exit 1;
fi

## Determine if we are running the 64 bit OS
## If sparcv9 then the location of the afs and nfs extensions differ

case `/bin/isalist` in
    *amd64* )
              nfssrv=/kernel/misc/amd64/nfssrv
              afs=/kernel/drv/amd64/afs ;;
    *sparcv9* )
              nfssrv=/kernel/misc/sparcv9/nfssrv
              afs=/kernel/drv/sparcv9/afs ;;
          * )
              nfssrv=/kernel/misc/nfssrv
              afs=/kernel/drv/afs ;;
esac


#
# Load kernel extensions
#
# nfssrv has to be loaded first


if [ -f $nfssrv ]; then
      echo "Loading NFS server kernel extensions"
      modload $nfssrv
else
      echo "$nfssrv does not exist. Skipping AFS startup."
      exit 1
fi

## Load AFS kernel extensions

if [ -f $afs ]; then
      if [ -f /kernel/drv/afs.conf ] ; then
          echo "Kernel afs.conf already exists"
      else
          echo "Creating kernel afs.conf"
          echo 'name="afs" parent="pseudo";' > /kernel/drv/afs.conf
      fi

      # load the module
      if grep '^afs ' /etc/name_to_major >/dev/null ; then
          echo "Loading AFS kernel extensions"
          modload $afs
	  # this can sometimes be necessary to get the /devices afs device to
	  # attach
	  update_drv afs
      else
          echo "Installing AFS driver and loading kernel extensions"
          add_drv -m '* 0666 root root' afs
      fi

      # Create the /dev/afs link
      if grep name=afs /etc/devlink.tab >/dev/null ; then
          echo "Entry for afs already exists in /etc/devlink.tab"
      else
          echo "Adding entry for afs in /etc/devlink.tab"
          echo "type=ddi_pseudo;name=afs;addr=0;minor=afs	\D" >> /etc/devlink.tab
	  devfsadm
      fi
else
      echo "$afs does not exist. Skipping AFS startup."
      exit 1
fi

#
# Check that all of the client configuration files exist
#
generate_csdb
for file in /usr/vice/etc/afsd /usr/vice/etc/cacheinfo \
          /usr/vice/etc/ThisCell /usr/vice/etc/CellServDB
do
      if [ ! -f ${file} ]; then
              echo "${file} does not exist. Not starting AFS client."
              exit 1
      fi
done

#
# Check that the root directory for AFS (/afs)
# and the cache directory (/usr/vice/cache) both exist
#

for dir in `awk -F: '{print $1, $2}' /usr/vice/etc/cacheinfo`
do
      if [ ! -d ${dir} ]; then
              echo "${dir} does not exist. Not starting AFS client."
              exit 2
      fi
done

echo "Starting afsd"
/usr/vice/etc/afsd $OPTIONS

echo ;;

'stop')

afsroot=`awk -F: '{print $1}' /usr/vice/etc/cacheinfo` || exit 1
if /usr/sbin/mount | grep -q "$afsroot"; then
    echo "Unmounting $afsroot"
    /usr/sbin/umount $afsroot || exit 1

    echo "Stopping afsd"
    /usr/vice/etc/afsd -shutdown || exit 1

    afsid=`modinfo | awk '/afs filesystem/ {print $1}'` || exit 1
    echo "Unloading module id $afsid"
    modunload -i "$afsid"  || exit 1
fi
echo ;;

*)    echo "Invalid option supplied to $0"
      exit 1;;
esac
""",


    "openafs-server.init":\
"""\
#!/bin/bash
#
# Copyright (c) 2014-2017, Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#----------------------------------------------------------------------
# Basic init script to start/stop the OpenAFS servers.
#----------------------------------------------------------------------
# chkconfig: 2345 49 51
#----------------------------------------------------------------------

if [ -f /etc/rc.d/init.d/functions ] ; then
    . /etc/rc.d/init.d/functions
    afs_rh=1
else
    # special (RedHat) functions not available...
    function echo_failure () { echo -n " - failed." ; }
    function echo_success () { echo -n " - successful." ; }
fi

is_on() {
    if  test "$1" = "on" ; then return 0
    else return 1
    fi
}

BOSSERVER_OPTIONS=""
SYSCNF=${SYSCNF:-/etc/sysconfig/openafs-server}
if [ -f $SYSCNF ] ; then
    . $SYSCNF
fi

BOS=${BOS:-/usr/afs/bin/bos}
BOSSERVER=${BOSSERVER:-/usr/afs/bin/bosserver}

start() {
    if [ ! "$afs_rh" -o ! -f /var/lock/subsys/openafs-server ]; then
        if test -x $BOSSERVER ; then
            echo "Starting AFS servers..... "
            $BOSSERVER $BOSSERVER_OPTIONS
            if [ $? -ne 0 ]; then
                echo "Failed to start bosserver!"
                exit 1
            fi
            test "$afs_rh" && touch /var/lock/subsys/openafs-server
            if is_on $WAIT_FOR_SALVAGE; then
                sleep 10
                while $BOS status localhost fs 2>&1 | grep 'Auxiliary.*salvaging'; do
                    echo "Waiting for salvager to finish..... "
                    sleep 10
                done
            fi
        fi
    fi
}

stop() {
    if [ ! "$afs_rh" -o -f /var/lock/subsys/openafs-server ]; then
        if  test -x $BOS ; then
            echo "Stopping AFS servers..... "
            $BOS shutdown localhost -localauth -wait
            pkill -HUP bosserver
        fi
        rm -f /var/lock/subsys/openafs-server
    fi
}

case "$1" in
  start)
      start
      ;;
  stop)
      stop
      ;;
  restart)
      $0 stop
      $0 start
      ;;
  *)
      echo $"Usage: $0 {start|stop|restart}"
      exit 1
esac
""",


}



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Common system utilities."""

import logging
import os
import subprocess
import sys

logger = logging.getLogger(__name__)

class RingBuffer:
    """Circular array for appending."""
    # Adapted from the python cookbook.
    def __init__(self, size_max):
        self.size_max = size_max
        self.data = []

    class __FullBuffer:
        def append(self, x):
            """Append an element overwriting the oldest one."""
            self.data[self.cursor] = x
            self.cursor = (self.cursor + 1) % self.size_max

        def get(self):
            """Return list of elements in correct order."""
            return self.data[self.cursor:] + self.data[:self.cursor]

    def append(self,x):
        """Append an element at the end of the buffer."""
        self.data.append(x)
        if len(self.data) == self.size_max:
            # We are full; switch to the circular functions.
            self.cursor = 0
            self.__class__ = self.__FullBuffer

    def get(self):
        """ Return a list of elements from the oldest to the newest. """
        return self.data


class CommandMissing(Exception):
    """Command not found."""

class CommandFailed(Exception):
    """Command exited with a non-zero exit code."""
    def __init__(self, args, code, out):
        self.cmd = subprocess.list2cmdline(args)
        self.args = args
        self.code = code
        self.out = out

    def __str__(self):
        msg = "Command failed! %s; code=%d, out='%s'" % \
              (self.cmd, self.code, self.out.strip())
        return repr(msg)

def sh(*args, **kwargs):
    """Execute the command line arguments.

    Execute the command line and optionally return the output as
    a list of lines.  Raises a CommandFailed exception if the
    command exits with a non-zero code.

    args:     command-line arguments
    output:   return output lines as a list (default: True)
    quiet:    do not log command line and output (default: False)
    prefix:   log message prefix (default: None)
    sed:      output line filter function (default: None)
    dryrun:   print the command instead of executing it (defualt: False)
    tailsize: number of lines to report when output=False (default:20)
    """
    output = kwargs.get('output', True)
    quiet = kwargs.get('quiet', False)
    prefix = kwargs.get('prefix', None)
    sed = kwargs.get('sed', None)
    dryrun = kwargs.get('dryrun', False)
    tailsize = kwargs.get('tailsize', 20)

    # Fixup the argument list for Popen.
    # 1. Create a tuple if just one arg was given.
    # 2. Convert numeric args to strings.
    if isinstance(args, basestring):
        args = (args)
    args = [arg.__str__() for arg in args]

    # Be sure the first arg is actually a program, otherwise Popen
    # will fail with a cryptic exception.
    args[0] = which(args[0], raise_errors=True)
    cmdline = subprocess.list2cmdline(args)

    # Dryrun mode: Just print what would be run.
    if dryrun:
        sys.stdout.write("%s\n" % (cmdline))
        return

    # Execute command and process output.
    lines = []
    if output and not sed:
        tail = None
    else:
        tail = RingBuffer(tailsize)  # Save the tail for error reporting.
    if quiet:
        if prefix:
            logger.debug("%s: running: %s", prefix, cmdline)
        else:
            logger.debug("running: %s", cmdline)
    else:
        if prefix:
            logger.info("%s: running: %s", prefix, cmdline)
        else:
            logger.info("running: %s", cmdline)
    p = subprocess.Popen(args,
                        bufsize=1,
                        env=os.environ,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT) # Redirect stderr capture errors.
    with p.stdout:
        for line in iter(p.stdout.readline, ''):
            line = line.rstrip("\n")
            if tail:
                tail.append(line)
            if not quiet:
                if prefix:
                    logger.info("%s: %s", prefix, line)
                else:
                    logger.info("%s", line)
            if output:
                if sed:
                    line = sed(line)
                if line:
                    lines.append(line)
    code = p.wait()
    if code != 0:
        if tail:
            lines = tail.get()
        out = "\n".join(lines)
        raise CommandFailed(args, code, out)
    return lines

def which(program, extra_paths=None, raise_errors=False):
    """Find a program in the PATH.

    program: program name or program full path
    extra_paths: list of paths to search in addition to PATH
    raise_errors: raise an exception if not found (default: False)
    """
    if not isinstance(program, basestring):
        raise ValueError("which() requires a string argument")
    dirname,basename = os.path.split(program)
    if dirname:
        # Full path was given; verify it is an executable file.
        if os.path.isfile(program) and os.access(program, os.X_OK):
            return program
        if raise_errors:
            raise CommandMissing("Program '%s' is not an executable file." % (program))
    else:
        # Just the basename was given; search the paths.
        paths = os.environ['PATH'].split(os.pathsep)
        if extra_paths:
            paths = paths + extra_paths
        for path in paths:
            path = path.strip('"')
            fpath = os.path.join(path, program)
            if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
                return fpath
        if raise_errors:
            raise CommandMissing("Could not find '%s' in paths %s" % (program, ":".join(paths)))
    return None

def file_should_exist(path, description=None):
    """Fails if the given file does not exist."""
    if not os.path.isfile(path):
        if description is None:
            description = "File '%s' does not exist." % (path)
        raise AssertionError(description)
    return True

def directory_should_exist(path, description=None):
    """Fails if the given directory does not exist."""
    if not os.path.isdir(path):
        if description is None:
            description = "Directory '%s' does not exist." % (path)
        raise AssertionError(description)
    return True

def directory_should_not_exist(path, description=None):
    """Fails if the given directory does exist."""
    if os.path.exists(path):
        if description is None:
            description = "Directory '%s' already exists." % (path)
        raise AssertionError(description)
    return True

def path_join(a, *p):
    # os.path.join() is brain dead.
    p = [x.lstrip('/') for x in p]
    return os.path.join(a, *p)

def nproc():
    """Return the number of processing units."""
    nproc = which('nproc')
    if nproc is None:
        return 1  # default
    return int(sh('nproc')[0])

def mkdirp(path):
    """Make a directory with parents."""
    # Do not raise an execption if the directory already exists.
    if not os.path.isdir(path):
        os.makedirs(path)

def cat(files, path):
    """Combine one or more files."""
    dst = open(path, 'w')
    for f in files:
        with open(f, 'r') as src:
            dst.write(src.read())
    dst.close()

def touch(path):
    """Touch a file; create a empty file if not already existing."""
    with open(path, 'a'):
        os.utime(path, None)

def symlink(src, dst):
    """Create a symlink dst to src."""
    logger.debug("Creating symlink %s -> %s", dst, src)
    if not os.path.isfile(src):
        raise AssertionError("Failed to make symlink: src %s not found" % (src))
    if os.path.exists(dst) and not os.path.islink(dst):
        raise AssertionError("Failed to make symlink: dst %s exists" % (dst))
    if os.path.islink(dst):
        os.remove(dst)
    os.symlink(src, dst)




from os import uname as _uname
if _uname()[0] == "Linux":
    from afsutil.system import linux as _mod
elif _uname()[0] == "SunOS":
    from afsutil.system import sunos as _mod
else:
    raise ImportError("Unsupported operating system.")

CommandMissing = _mod.CommandMissing
CommandFailed = _mod.CommandFailed
afs_mountpoint = _mod.afs_mountpoint
afs_umount = _mod.afs_umount
cat = _mod.cat
configure_dynamic_linker = _mod.configure_dynamic_linker
detect_gfind = _mod.detect_gfind
directory_should_exist = _mod.directory_should_exist
directory_should_not_exist = _mod.directory_should_not_exist
file_should_exist = _mod.file_should_exist
get_running = _mod.get_running
is_afs_mounted = _mod.is_afs_mounted
is_loaded = _mod.is_loaded
is_running = _mod.is_running
mkdirp = _mod.mkdirp
network_interfaces = _mod.network_interfaces
nproc = _mod.nproc
path_join = _mod.path_join
sh = _mod.sh
symlink = _mod.symlink
tar = _mod.tar
touch = _mod.touch
unload_module = _mod.unload_module
untar = _mod.untar
which = _mod.which



# Copyright (c) 2014-2016 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Wrappers to invoke AFS command line tools."""

import os
import logging
import time

from afsutil.system import sh, which, CommandFailed
from afsutil.transarc import AFS_SRV_BIN_DIR, AFS_SRV_SBIN_DIR, AFS_WS_DIR

logger = logging.getLogger(__name__)

# Common hidding places.
PATHS = [
    '/usr/bin',
    '/usr/sbin',
    AFS_SRV_BIN_DIR,
    AFS_SRV_SBIN_DIR,
    os.path.join(AFS_WS_DIR, 'bin'),
    os.path.join(AFS_WS_DIR, 'etc'),
    '/usr/kerberos/bin',
    '/usr/heimdal/bin',
]

_cmdpath = {
}

def setpath(cmd, path):
    _cmdpath[cmd] = path

def _run(cmd, args=None, quiet=False, retry=0, wait=1, cleanup=None):
    """Execute a command and return the output as a string.

    cmd:     command to be executed
    args:    list of command line arguments
    quiet:   do not log command and output
    retry:   number of retry attempts, 0 for none
    wait:    delay between retry attempts
    cleanup: cleanup function to run before retry

    returns: command output as a string

    Raises a CommandFailed exception if the command exits with
    a non-zero exit code."""
    count = 0 # retry counter
    if args is None:
        args = []
    elif not isinstance(args, list):
        args = list(args)
    cmd = _cmdpath.get(cmd, cmd)
    args.insert(0, which(cmd, raise_errors=True, extra_paths=PATHS))
    while True:
        try:
            lines = sh(*args, quite=quiet)
            break
        except CommandFailed as cf:
            if count < retry:
                count += 1
                logger.info("Retrying %s command in %d seconds; retry %d of %d.",
                    cmd, wait, count, retry)
                time.sleep(wait)
                if cleanup:
                    cleanup()  # Try to cleanup the mess from the last failure.
            else:
                raise cf
    return "\n".join(lines)

def asetkey(*args, **kwargs):
    return _run('asetkey', args=args, **kwargs)

def kinit(*args, **kwargs):
    return _run('kinit', args=args, **kwargs)

def aklog(*args, **kwargs):
    return _run('aklog', args=args, **kwargs)

def bos(*args, **kwargs):
    if os.geteuid() == 0:
        args = list(args)
        args.append('-localauth')
    return _run('bos', args=args, **kwargs)

def vos(*args, **kwargs):
    if os.geteuid() == 0:
        args = list(args)
        args.append('-localauth')
    return _run('vos', args=args, **kwargs)

def pts(*args, **kwargs):
    if os.geteuid() == 0:
        args = list(args)
        args.append('-localauth')
    return _run('pts', args=args, **kwargs)

def fs(*args, **kwargs):
    return _run('fs', args=args, **kwargs)

def udebug(*args, **kwargs):
    return _run('udebug', args=args, **kwargs)

def rxdebug(*args, **kwargs):
    return _run('rxdebug', args=args, **kwargs)

def tokens(*args, **kwargs):
    return _run('tokens', args=args, **kwargs)



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""System checks"""

import logging
import re
import socket
import sys
from afsutil.system import network_interfaces

logger = logging.getLogger(__name__)

def check_hosts_file():
    """Check for loopback addresses in the /etc/hosts file.

    Modern linux distros assign a loopback address to the hostname. While not
    strictly incorrect, this can confuse OpenAFS which gets our IP address from
    gethostbyname()."""
    result = True
    nr = 0
    hostname = socket.gethostname()
    with open('/etc/hosts', 'r') as f:
        for line in f.readlines():
            nr += 1
            if line.startswith('127.'):
                if hostname in line.split()[1:]:
                    sys.stderr.write(
                        "Warning: loopback address '%s' assigned to our hostname '%s' on line %d of /etc/hosts.\n" % \
                        (line.split()[0], hostname, nr))
                    result = False
    return result

def fix_hosts_file():
    hostname = socket.gethostname()
    ips = network_interfaces()
    if len(ips) == 0:
        raise AssertionError("Unable to detect any non-loopback network interfaces.")
    ip = ips[0]

    with open('/etc/hosts', 'r') as f:
        hosts = f.read()

    with open('/etc/hosts', 'w') as f:
        for line in hosts.splitlines():
            line = line.strip()
            if not re.search(r'\b%s\b' % (hostname), line):
                f.write("%s\n" % line)
        f.write("%s\t%s\n" % (ip, hostname))

def check_host_address():
    """Verify our hostname resolves to a useable address."""
    hostname = socket.gethostname()
    ip = socket.gethostbyname(hostname)
    ips = network_interfaces()
    if len(ips) == 0:
        sys.stderr.write("Warning: Unable to detect any non-loopback network interfaces.\n")
        return False
    if not ip in ips:
        sys.stderr.write("Warning: hostname '%s' resolves to '%s', not found in network interfaces: '%s'.\n" % \
                         (hostname, ip, " ".join(ips)))
        return False
    return True

def check(**kwargs):
    result = 0
    if not check_hosts_file():
        if kwargs['fix_hosts']:
            sys.stdout.write("Attempting to fix /etc/hosts file.\n")
            fix_hosts_file()
        else:
            result = 1
    if not check_host_address():
        result = 2
    return result



# Copyright (c) 2014-2019 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Driver to build OpenAFS rpms from git sources

This module provides classes to build OpenAFS srpms, client and server
userspace rpms, and linux kernel module rpms. Mock may be used to build in a
chroot, which is the preferred method since build dependencies will be
contained to the temporary chroot.

The RpmBuilder searches for linux kernel-devel packages installed on the system
and will build an OpenAFS kmod package for each linux version found. Be sure
your system has a kernel-devel package installed and is up to date to build a
kmod for your running kernel version.

The MockRpmBuilder searches for linux kernel-devel packages available in the
chroot by querying all the enabled yum repos inside the chroot. The
MockRpmBuilder will build an OpenAFS kmod package for each available
kernel-devel package. The chroot can be configured to have private yum repos,
which can be used to provide a large number of kernel-devel versions.

Before building the rpms, clone the OpenAFS git repo and check out the tag to
be built. For example:

    $ cd $HOME
    $ git clone https://github.com/openafs/openafs.git
    $ cd openafs
    $ git checkout openafs-stable-1_6_20

Example usage:

    from os.path import expanduser
    from afsutil.package import RpmBuilder

    b = RpmBuilder(srcdir=expanduser("~/openafs"))
    b.prepare_sources()
    b.build_srpm()
    b.build_userspace()
    b.build_kmods()

"""
from __future__ import print_function
import os
import sys
import re
import urllib2
import logging
import shutil
import glob
from afsutil.system import sh, mkdirp, which, CommandFailed
from afsutil.misc import flatten, trim

logger = logging.getLogger(__name__)

class RpmBuilderError(Exception):
    """Failed to build packages."""

def readfile(path):
    """Read a file into a string."""
    with open(path, 'r') as f:
        return f.read()

def writefile(path, contents):
    """Write string to a file."""
    with open(path, 'w') as f:
        f.write(contents)

class RpmBuilder(object):
    def __init__(self, srcdir=None, pkgdir=None, topdir=None, dstdir=None,
                 version=None, arch=None, spec=None, csdb=None, srpm=None,
                 clobber=False, quiet=False, verbose=False,
                 with_=None, without=None, **kwargs):
        """Initialize the RpmBuilder object

        srcdir:  path of the checked out source tree (default: .)
        pkgdir:  path of the packaging files (default: {srcdir}/src/packaging/RedHat
        topdir:  path of the rpmbuild directories (default: {srcdir}/packaging/rpmbuild)
        dstdir:  path to place rpms (default: {topdir}/RPMS/{arch})
        version: target version number (default: check)
        arch:    target architecture (default: check)
        spec:    custom rpmbuild specfile (default: {pkdir}/openafs.spec.in}
        srpm:    path of a prebuilt src.rpm file
        clobber: build and overwrite existing kmod-openafs rpms
        quiet:   less output
        verbose: more output
        with_:   rpmbuild --with options
        without: rpmbuild --without options
        """
        if srcdir is None:
            srcdir = os.getcwd()
        if pkgdir is None:
            pkgdir = os.path.join(srcdir, "src/packaging/RedHat")
        if topdir is None:
            topdir = os.path.join(srcdir, "packages/rpmbuild")
        if arch is None:
            arch = os.uname()[4]
        if with_ is None:
            with_ = []
        if without is None:
            without = []
        if dstdir is None:
            dstdir = os.path.join(topdir, "RPMS", arch)
        # paths
        self.srcdir = srcdir
        self.pkgdir = pkgdir
        self.topdir = topdir
        self.dstdir = dstdir
        # options
        self.clobber = clobber
        self.quiet = quiet
        self.verbose = verbose
        self.custom_spec = spec
        self.custom_csdb = csdb
        self.with_ = flatten(with_)
        self.without = flatten(without)
        # state
        self.version = version
        self.arch = arch
        self.spec = None
        self.csdb = None
        self.sources = {}
        self.srpm = srpm
        self.downloaded = []
        self.generated = []
        self.skipped = []
        self.failed = []
        self.built = []

    def withargs(self):
        """Return a list of optional --with and --without options."""
        args = []
        for with_ in self.with_:
            args.append('--with')
            args.append(with_)
        for without in self.without:
            args.append('--without')
            args.append(without)
        return args

    def rpmbuild(self, *args, **kwargs):
        """Run the rpmbuild commands."""
        return sh(
            'rpmbuild',
            '--define', '_topdir {0}'.format(self.topdir),
            *args, **kwargs)

    def git(self, *args):
        """Helper to run git commands in the source tree."""
        if not self.srcdir or self.srcdir == ".":
            gitdir = None
        else:
            gitdir = os.path.join(self.srcdir, ".git")
        args = list(args)
        args.insert(0, 'git')
        if gitdir:
            args.insert(1, "--git-dir")
            args.insert(2, gitdir)
        return sh(*args, quiet=True)

    def get_version(self):
        """Get the version identifier for the packaging.

        Use the verision given in the .version file if present. Otherwise
        try to use git to find the version number of the checked out tree.
        """
        if self.version:
            return self.version
        dot_version = os.path.join(self.srcdir, ".version")
        version = None
        if os.path.exists(dot_version):
            # Note: If the .version file exists, the value of the working file is
            # used when populating SOURCES, not the one extracted from git.
            logger.debug("getting OpenAFS version number from .version file.")
            version = readfile(dot_version).strip()
        else:
            logger.debug("getting OpenAFS version number with git describe.")
            output = self.git('describe', '--abbrev=4', 'HEAD')
            if output:
                version = output[0]
        if not version:
            raise RpmBuilderError("Failed to find version number.")
        self.version = re.sub(r'^openafs-[^-]*-', '', version).replace('_', '.')
        logger.info("OpenAFS version is {0}".format(self.version))
        return self.version

    def get_package_version(self, version=None):
        """Determine the packaging release info.

        The format of the linux packaging release and version strings depend on
        the type of release. The following variants are supported:

        * stardard release; 1.6.20 => 1.6.20, 1
        * prereleases; 1.6.20pre1 => 1.6.20, 0.pre1
        * development release; 1.8.0dev => 1.8.0, 0.dev
        * not tagged; e.g., 1.6.0-32-gabcdef => 1.6.0, .32.gabcdef
        """
        if version is None:
            version = self.get_version()
        m1 = re.match(r'(.*)(pre[0-9]+)', version) # prerelease
        m2 = re.match(r'(.*)dev', version) # development
        m3 = re.match(r'(.*)-([0-9]+)-(g[a-f0-9]+)$', version) # development
        m4 = re.match(r'(.*)-([a-z]+)([0-9]+)', version) # custom
        if m1:
            linux_pkgver = m1.group(1)
            linux_pkgrel = "0.{0}".format(m1.group(2))
        elif m2:
            linux_pkgver = m2.group(1);
            linux_pkgrel = "0.dev"
        elif m3:
            linux_pkgver = m3.group(1)
            linux_pkgrel = "{0}.{1}".format(m3.group(2),m3.group(3))
        elif m4:
            linux_pkgver = m4.group(1)
            linux_pkgrel = "1.2.{0}.{1}".format(m4.group(3), m4.group(2))
        else:
            linux_pkgver = version # standard release
            linux_pkgrel = "1"     # increment when repackaging this version
        logger.info("Linux version is {0}".format(linux_pkgver))
        logger.info("Linux release is {0}".format(linux_pkgrel))
        return (linux_pkgver, linux_pkgrel)

    def current_kversion(self):
        """Return the version of the running kernel."""
        version = os.uname()[2]
        arch = os.uname()[4]
        return trim(version, '.'+arch)

    def find_kversions(self):
        """Find the linux kernel versions of the kernel-devel packages."""
        kversions = sh(
            'rpm', '-q', '-a',
            '--queryformat=%{VERSION}-%{RELEASE}\n', 'kernel-devel',
            quiet=True)
        logger.debug("Found kernel versions: {0}".format(", ".join(kversions)))
        return kversions

    def find_kmods(self):
        """Find the linux kernel versions of the kmod rpms previously built.

        Find the linux kernel versions of the kmods which have been already
        built for this version of OpenAFS. This allows us to avoid rebuilding
        kmods for a given OpenAFS version when a new kernel-devel version
        becomes available.
        """
        if self.dstdir is None or not os.path.exists(self.dstdir):
            return []
        logger.info("Checking for existing kmods in {0}".format(self.dstdir))
        linux_pkgver,linux_pkgrel = self.get_package_version()
        logger.debug("linux_pkgver={0}, linux_pkgrel={1}".format(linux_pkgver,linux_pkgrel))
        def get_kversion(line):
            # Extract the linux kernel version from the rpm RELEASE tag.
            # e.g. 1.3.10.0_514.10.2.el7 => 3.10.0-514.10.2.el7
            v,r = line.split(' ', 2)
            prefix = linux_pkgrel + '.'
            if v == linux_pkgver and r.startswith(prefix):
                return r.replace(prefix, '', 1).replace('_', '-')
            return None
        kversions = sh(
            'rpm', '-q', '-p', '--queryformat=%{VERSION} %{RELEASE}\n',
            "{dstdir}/kmod-openafs*.rpm".format(dstdir=self.dstdir),
            sed=get_kversion)
        logger.debug("Found kmods: {0}".format(", ".join(kversions)))
        return kversions

    def generate_spec(self, src):
        """Helper to generate the openafs.spec file."""
        version = self.get_version()
        (linux_pkgver, linux_pkgrel) = self.get_package_version(version)
        dst = os.path.join(self.topdir, "SPECS", "openafs.spec")
        logger.info("Generating spec '{0}' from template '{1}'.".format(dst, src))
        mkdirp(os.path.dirname(dst))
        writefile(dst,
            readfile(src).\
               replace('@VERSION@', version).\
               replace('@PACKAGE_VERSION@', version).\
               replace('@LINUX_PKGVER@', linux_pkgver).\
               replace('@LINUX_PKGREL@', linux_pkgrel))
        self.generated.append(dst)
        return dst

    def copy_spec(self, src):
        """Helper to copy the spec file to the SPECS directory."""
        dst = os.path.join(self.topdir, "SPECS", "openafs.spec")
        logger.debug("Copying {0} to {1}".format(src, dst))
        mkdirp(os.path.dirname(dst))
        shutil.copy(src, dst)
        return dst

    def prepare_spec(self):
        """Copy the specfile to the SPECS directory.

        By default, generate the openafs.spec file from the openafs.spec.in
        template found in the packaging files directory in the source tree.
        Otherwise, copy the specfile given to the SPECS directory.

        """
        if self.spec:
            return # already have a specfile
        if self.custom_spec is None:
            spec = os.path.join(self.pkgdir, "openafs.spec.in")
        else:
            spec = self.custom_spec
        if spec.endswith(".in"):
            self.spec = self.generate_spec(spec)
        else:
            self.spec = self.copy_spec(spec)

    def list_sources(self, spec=None):
        """Helper to list the source tags in a specfile."""
        if spec is None:
            if self.spec is None:
                self.prepare_spec()
            spec = self.spec
        sources = []
        if spec is None:
            spec = os.path.join(self.topdir, "SPECS", "openafs.spec")
        for line in readfile(spec).splitlines():
            m = re.match(r'Source([0-9]*): (.*)', line)
            if m:
                n = m.group(1)
                v = m.group(2).replace('%{afsvers}', self.version)
                sources.append((n,v))
        return sources

    def lookup_source(self, pattern, spec=None):
        """Helper to find a source tag in the specfile."""
        for n,source in self.list_sources(spec=spec):
            if re.search(pattern, source):
                return (n,source)
        return (None, None)

    def download_csdb(self, url):
        """Helper to download the CellServDB file to SOURCES."""
        dst = os.path.join(self.topdir, 'SOURCES', os.path.basename(url))
        logger.info("Downloading CellServDB from '{0}' to {1}'".format(url, dst))
        src = urllib2.urlopen(url)
        mkdirp(os.path.dirname(dst))
        writefile(dst, src.read())
        self.downloaded.append(dst)
        return dst

    def copy_csdb(self, src):
        """Helper to copy the CellServDB file to SOURCES."""
        dst = os.path.join(self.topdir, 'SOURCES', os.path.basename(src))
        logger.info("Copying user supplied CellServDB '{0}' to '{1}'.".format(src, dst))
        mkdirp(os.path.dirname(dst))
        shutil.copy(src, dst)
        self.downloaded.append(dst)
        return dst

    def change_csdb_source(self, csdb):
        n,v = self.lookup_source('CellServDB')
        old = "Source{0}: {1}".format(n, v)
        new = "Source{0}: {1}".format(n, os.path.basename(csdb))
        logger.info("Updating spec {0} with new CellServDB name.".format(self.spec))
        writefile(self.spec, readfile(self.spec).replace(old, new))

    def prepare_sources_csdb(self):
        """Set the CellServDB file to be packaged.

        Sets the CellServDB file to be packaged. This file will be copied to
        /usr/vice/etc/CellServDB.dist during packaging.  Call this function
        after prepare_spec() and before prepare_sources() to set a custom
        CellServDB file.
        """
        if self.csdb:
            return # already have the CellServDB
        if self.custom_csdb:
            csdb = self.custom_csdb
            if csdb.startswith('http://') or csdb.startswith('https://'):
                self.csdb = self.download_csdb(csdb)
            else:
                self.csdb = self.copy_csdb(csdb)
            if self.spec:
                self.change_csdb_source(csdb)
            else:
                logger.warning("Skipping csdb source update in spec file.")
        else:
            n,url = self.lookup_source(r'^https?://.*CellServDB')
            if url:
                self.csdb = self.download_csdb(url)
            else:
                raise RpmBuilderError("CellServDB url not found in spec file.""")

    def prepare_sources_relnotes(self):
        """Create the release notes to be packaged.

        Extract the changes for this version from the running NEWS file in the
        source tree.
        """
        if 'relnotes' in self.sources:
            return # already have release notes
        version = self.get_version()
        infile = os.path.join(self.srcdir, "NEWS")
        outfile = os.path.join(self.topdir, "SOURCES", "RELNOTES-{0}".format(version))
        news = readfile(infile).splitlines()
        try:
            i = news.index("OpenAFS {0}".format(version))
        except ValueError:
            # NEWS not updated for this release? Provide a file
            # to keep rpmbuild happy since this is a required SOURCE.
            relnotes = []
        else:
            # Extact the news items for just this release.
            relnotes = news[0:i+1] # the preamble
            for line in news[i+1:]:
                if line.startswith("OpenAFS"):
                    break # start of next section
                relnotes.append(line)
        mkdirp(os.path.dirname(outfile))
        writefile(outfile, "\n".join(relnotes))
        self.generated.append(outfile)
        self.sources['relnotes'] = True

    def prepare_sources_changelog(self):
        """Generate the ChangeLog file from git log.

        Use the git log as the ChangeLog.
        """
        if 'changelog' in self.sources:
            return
        dst = "{topdir}/SOURCES/ChangeLog".format(topdir=self.topdir)
        mkdirp(os.path.dirname(dst))
        cmd = "git --git-dir={srcdir}/.git log > {dst}". format(srcdir=self.srcdir, dst=dst)
        logger.info("running {cmd}".format(cmd=cmd))
        os.system(cmd)
        self.generated.append(dst)
        self.sources['changelog'] = True

    def prepare_sources_scripts(self):
        """Copy packaging helper scripts to the SOURCES directory.

        Copy packaging scripts listed in the spec file from the source tree
        into the rpmbuild SOURCES directory.
        """
        if 'scripts' in self.sources:
            return
        if not os.path.isdir(self.pkgdir):
            raise RpmBuilderError("Path to packaging scripts not found: '{0}'.".format(self.pkgdir))
        for n,source in self.list_sources():
            src = os.path.join(self.pkgdir, os.path.basename(source))
            if os.path.isfile(src):
                dst = os.path.join(self.topdir, "SOURCES", os.path.basename(source))
                logger.debug("Copying file '{0}' to '{1}'.".format(src, dst))
                mkdirp(os.path.dirname(dst))
                shutil.copy(src, dst)
                if dst.endswith(".sh"):
                    os.chmod(dst, 0755)
        self.sources['scripts'] = True

    def prepare_sources_tarballs(self):
        """Create the source and doc tarballs from git sources.

        Extract the source tree from the git repo and create the src and doc
        tarball files. The tarballs will be placed into the rpmbuild SOURCES
        directory.
        """
        if 'tarballs' in self.sources:
            return
        version = self.get_version()
        names = dict(
            version=version,
            topdir=self.topdir,
        )
        tarfile = "{topdir}/SOURCES/openafs-{version}.tar".format(**names)

        logger.info("Extracting source tree from git to {0}".format(os.path.dirname(tarfile)))
        mkdirp(os.path.dirname(tarfile))
        self.git('archive', '--format', 'tar',
                 '--prefix', 'openafs-{version}/'.format(**names),
                 '--output', tarfile,
                 'HEAD')
        sh('tar', 'xf', tarfile, '-C', os.path.dirname(tarfile), output=False)
        os.remove(tarfile)

        logger.info("Generating source tree and documents.")
        # Note: Overwrite the extracted version string, if one.
        writefile('{topdir}/SOURCES/openafs-{version}/.version'.format(**names), version)
        sh('/bin/sh', '-c',
            'cd {topdir}/SOURCES/openafs-{version} && ./regen.sh'.format(**names),
            output=False)

        logger.info("Creating doc tarball openafs-{version}-doc.tar.bz2".format(**names))
        sh('tar', 'cjf',
           '{topdir}/SOURCES/openafs-{version}-doc.tar.bz2'.format(**names),
           '-C', '{topdir}/SOURCES'.format(**names),
           'openafs-{version}/doc'.format(**names),
           output=False)
        self.generated.append('{topdir}/SOURCES/openafs-{version}-doc.tar.bz2'.format(**names))

        logger.info("Creating src tarball openafs-{version}-src.tar.bz2".format(**names))
        sh('tar', 'cjf',
           '{topdir}/SOURCES/openafs-{version}-src.tar.bz2'.format(**names),
           '-C', '{topdir}/SOURCES'.format(**names),
           '--exclude', 'doc',
           'openafs-{version}'.format(**names),
           output=False)
        self.generated.append('{topdir}/SOURCES/openafs-{version}-src.tar.bz2'.format(**names))
        # Clean up the temporary tree.
        shutil.rmtree('{topdir}/SOURCES/openafs-{version}'.format(**names))
        self.sources['tarballs'] = True

    def prepare_sources(self):
        """Prepare the rpmbuild SOURCES directory.

        arguments:
        spec: custom rpmbuild specfile
        csdb: custom CellServDB file to package
        """
        if 'all' in self.sources:
            return
        self.banner(["Preparing sources"])
        self.prepare_spec()
        self.prepare_sources_csdb()
        self.prepare_sources_relnotes()
        self.prepare_sources_changelog()
        self.prepare_sources_scripts()
        self.prepare_sources_tarballs()
        self.sources['all'] = True

    def build_srpm(self):
        """Build the source rpm.

        Run rpmbuild to build the source rpm from the source tarball and other
        files in SOURCES. The rpmbuild SOURCE and SPEC directories must be
        populated with the prepare_* functions before calling build_srpm.
        """
        self.prepare_spec()
        self.prepare_sources()
        self.banner(["Building srpm"])
        def name_written(line):
            m = re.match(r'Wrote: (.*\.src\.rpm)$', line)
            return m.group(1) if m else None
        args = self.withargs()
        args.append(self.spec)
        output = self.rpmbuild(
            '-bs', '--nodeps',
            '--define', 'build_userspace 1',
            '--define', 'build_modules 0',
            *args,
            sed=name_written)
        if len(output) < 1:
            raise RpmBuilderError("Failed to get srpm name.")
        self.srpm = output[0]
        self.built.append(self.srpm)
        logger.info("srpm is {0}".format(self.srpm))

    def build_userspace(self, srpm=None):
        """Build the userspace rpms.

        Run rpmbuild to build the userspace biniary rpms from the source rpm.

        srpm: path of the srpm. defaults to the srpm created by the
              previous build_srpm()
        """
        if srpm is None:
            if self.srpm is None:
                self.build_srpm()
            srpm = self.srpm

        self.banner(["Building rpms"])
        def get_wrote(line):
            m = re.match(r'Wrote: (.*\.rpm)$', line)
            return m.group(1) if m else None
        args = self.withargs()
        args.append(srpm)
        output = self.rpmbuild(
            '--rebuild', '-bb',
            '--target', self.arch,
            '--define', 'build_userspace 1',
            '--define', 'build_modules 0',
            *args,
            sed=get_wrote)
        self.built.extend(output)

    def build_kmod(self, srpm=None, kversion=None):
        """Build the a kmod rpm for the given kernel version.

        srpm:     path of the source rpm file
                  defaults to the srpm from the previous build_srpm()
        kversion: target linux kernel version, e.g., 3.10.0-514.16.1.el7
                  defaults to the currently running linux kernel version.

        A kernel-devel package for `kversion` must been installed to supply the
        required linux kernel header files.
        """
        if srpm is None:
            if self.srpm is None:
                self.build_srpm()
            srpm = self.srpm

        if kversion is None:
            kversion = self.current_kversion()
        else:
            kversion = trim(kversion, '.'+self.arch)

        info = ["Building module for {kversion}".format(kversion=kversion)]
        self.banner(info)

        logger.info("Building kmod for linux version {0}.".format(kversion))
        def name_written(line):
            m = re.match(r'Wrote: (.*\.rpm)$', line)
            return m.group(1) if m else None
        output = self.rpmbuild(
            '--rebuild', '-bb',
            '--target', self.arch,
            '--define', 'kernvers {0}'.format(kversion),
            '--define', 'build_userspace 0',
            '--define', 'build_modules 1',
            srpm,
            sed=name_written)
        self.built.extend(output)

    def build_kmods(self, srpm=None, kversions=None):
        """Build kmods for each kernel versions.

        srpm:      path of the source rpm file
                   defaults to the srpm from the previous build_srpm()
        kversions: list of target kernel versions
                   defaults to the list of versions of currently
                   installed kernel-devel packages.
        """
        if not kversions:
            kversions = self.find_kversions()
        kmods = self.find_kmods()  # previously built
        for kversion in kversions:
            if kversion in kmods:
                logger.info('Skipping already built kmod for version %s', kversion)
                self.skipped.append(kversion)
                continue
            try:
                self.build_kmod(srpm=srpm, kversion=kversion)
            except CommandFailed as fail:
                self.failed.append(kversion)
                logger.warning("Failed to build kmod for version {0}".format(kversion))
                logger.exception(fail)

    def createrepo(self):
        """Run createrepo in the destination directory."""
        if self.dstdir and os.path.exists(self.dstdir):
            createrepo = which('createrepo')
            if not createrepo:
                logger.warning("createrepo is not installed.")
            else:
                sh('createrepo', self.dstdir, output=False)

    def banner(self, lines):
        """Print a banner."""
        if not self.quiet:
            bar = "*" * 72 + "\n"
            sys.stdout.write(bar)
            for line in flatten(lines):
                sys.stdout.write("*  {0}\n".format(line))
            sys.stdout.write(bar)

    def summary(self):
        """Display summary of results."""
        summary = ["Summary", ""]
        for name in ('downloaded', 'generated', 'skipped', 'built', 'failed'):
            items = getattr(self, name)
            if items:
                summary.append(name.capitalize() + ':')
                summary.append(items)
                summary.append('')
        self.banner(summary)

class MockRpmBuilder(RpmBuilder):
    def __init__(self, chroot, autoclean=True, khrepo='kernel-devel', **kwargs):
        """Initialize the MockRpmBuilder object.

        chroot: name of the chroot (mock --root)
        """
        if chroot is None:
            raise ValueError("chroot argument is required.")
        self.chroot = chroot
        self.khrepo = khrepo # may be None
        self.autoclean = autoclean
        self.inited = False
        RpmBuilder.__init__(self, **kwargs)

    def mock(self, *args, **kwargs):
        args = list(args)
        if self.verbose:
            args.append('--verbose')
        return sh('mock', '--root', self.chroot, *args, **kwargs)

    def init_chroot(self):
        """Initialize the chroot."""
        if not self.inited:
            logger.info("Initializing chroot {0}.".format(self.chroot))
            self.mock('--init', '--quiet', output=False)
            self.inited = True

    def __del__(self):
        """Remove the chroot."""
        if self.inited and self.autoclean:
            logger.info("Removing chroot {0}.".format(self.chroot))
            self.mock('--clean', '--quiet', output=False)
            self.inited = False

    def find_kversions(self):
        """List the linux kernel versions of the available kernel header packages in the chroot."""
        self.init_chroot()
        self.mock('--install', 'yum-utils', '--quiet', output=False) # for repoquery
        cmd = ['repoquery']
        if self.khrepo:
            cmd.append('--repoid={0}'.format(self.khrepo))
        cmd.append('--show-dupes')
        cmd.append('--queryformat="%{VERSION}-%{RELEASE}"')
        cmd.append('kernel-devel')
        if self.khrepo:
            logger.debug('Searching for kernel-devel in repo %s.', self.khrepo)
        else:
            logger.debug('Searching for kernel-devel in enabled repos.')
        output = self.mock('--quiet', '--chroot', ' '.join(cmd), quiet=True)
        versions = []
        for line in output:
            if line.startswith('rpmdb:'):
                continue # skip error messages
            if line.startswith('error:'):
                continue # skip error messages
            logger.debug("found kernel-devel for kernel version: %s", line)
            versions.append(line)
        return versions

    def build_srpm(self):
        """Build the source rpm with mock."""
        self.prepare_spec()
        self.prepare_sources()
        self.banner(["Building srpm"])
        resultdir = "/var/lib/mock/{chroot}/result".format(chroot=self.chroot)
        self.mock(
            '--buildsrpm',
            '--resultdir', resultdir,
            '--spec', self.spec,
            '--sources', '{topdir}/SOURCES'.format(topdir=self.topdir),
            output=False)
        rpms = glob.glob("{resultdir}/*.src.rpm".format(resultdir=resultdir))
        if len(rpms) < 1:
            raise RpmBuilderError("Failed to get srpm name.")
        src = rpms[0]
        dst = os.path.join(self.topdir, "SRPMS", os.path.basename(src))
        mkdirp(os.path.dirname(dst))
        shutil.copy(src, dst)
        self.srpm = dst
        self.built.append(self.srpm)

    def build_userspace(self, srpm=None):
        """Build the userspace rpms using mock."""
        if srpm is None:
            if self.srpm is None:
                self.build_srpm()
            srpm = self.srpm

        self.banner(["Building rpms"])
        self.init_chroot()
        logger.info("Building userspace rpms in chroot {chroot}".format(chroot=self.chroot))
        resultdir = "/var/lib/mock/{chroot}/result".format(chroot=self.chroot)
        args = self.withargs()
        args.append(srpm)
        self.mock(
            '--rebuild',
            '--arch', self.arch,
            '--resultdir', resultdir,
            '--define', 'build_userspace 1',
            '--define', 'build_modules 0',
            *args,
            output=False)
        for rpm in glob.glob("{resultdir}/*.rpm".format(resultdir=resultdir)):
            mkdirp(self.dstdir)
            dst = "{dstdir}/{rpm}".format(dstdir=self.dstdir, rpm=os.path.basename(rpm))
            self.built.append(dst)
            shutil.copy(rpm, dst)
            logger.info("Wrote: {dst}".format(dst=dst))

    def build_kmod(self, srpm=None, kversion=None):
        """Build the a kmod rpm for the given kernel version.

        srpm:     path of the source rpm file
                  defaults to the srpm from the previous build_srpm()
        kversion: target linux kernel version, e.g., 3.10.0-514.16.1.el7
                  defaults to the currently running linux kernel version.

        A kernel-devel package for `kversion` must been installed to supply the
        required linux kernel header files.
        """
        if srpm is None:
            if self.srpm is None:
                self.build_srpm()
            srpm = self.srpm
        if kversion is None:
            kversion = self.current_kversion()
        else:
            kversion = trim(kversion, '.'+self.arch)

        info = ["Building module for {kversion}".format(kversion=kversion)]
        self.banner(info)

        self.init_chroot()
        logger.info("Building kmod for linux version {0}.".format(kversion))
        resultdir = "/var/lib/mock/{chroot}/result".format(chroot=self.chroot)
        args = self.withargs()
        args.append(srpm)
        self.mock(
            '--rebuild',
            '--arch', self.arch,
            '--resultdir', resultdir,
            '--define', 'kernvers {0}'.format(kversion),
            '--define', 'build_userspace 0',
            '--define', 'build_modules 1',
            *args,
            output=False)
        for rpm in glob.glob("{resultdir}/*.rpm".format(resultdir=resultdir)):
            mkdirp(self.dstdir)
            dst = "{dstdir}/{rpm}".format(dstdir=self.dstdir, rpm=os.path.basename(rpm))
            self.built.append(dst)
            shutil.copy(rpm, dst)
            logger.info("Wrote: {dst}".format(dst=dst))

def package(**kwargs):
    """Build OpenAFS rpms.

    keyword arguments:
    chroot: name of the mock chroot (default: dont use mock)
    build: what to build: 'all', 'srpm', 'userspace', 'kmods' (default: 'all')
    spec: custom rpmbuild specfile (default: in-tree openafs.spec.in)
    csdb: custom CellServDB file (default: download from url in specfile)
    srpm: path to a prebuilt src.rpm (default: build from sources)
    srcdir: path of the checked out source tree (default: .)
    pkgdir: path of the packaging files (default: {srcdir}/src/packaging/RedHat
    topdir: path of the rpmbuild directories (default: {srcdir}/packaging/rpmbuild)
    dstdir: path to place rpms (default: {topdir}/RPMS/{arch})
    arch: target arch (default: check)
    clobber: build and overwrite existing kmod-openafs rpms
    quiet:   less output
    """
    # Note: It could be possible no kmods need to be built, in which case it it
    # not neccessary to build the srpm. So we defer prepare_sources until we
    # know it is needed.
    chroot = kwargs.pop('chroot', None)
    build = kwargs.pop('build', 'all')
    kversions = flatten(kwargs.pop('kversions', None)) # argparse gives a list of lists
    list_kversions = kwargs.get('list_kversions')
    dstdir = kwargs.get('dstdir')

    if chroot:
        b = MockRpmBuilder(chroot, **kwargs)
    else:
        b = RpmBuilder(**kwargs)

    if list_kversions:
        kversions = b.find_kversions()
        for kversion in kversions:
            print(kversion)
        return

    if build == 'sources':
        b.prepare_spec()
        b.prepare_sources()
    elif build == 'srpm':
        b.build_srpm()
    elif build == 'userspace':
        b.build_userspace()
        if dstdir:
            b.createrepo()
    elif build == 'kmods':
        b.build_kmods(kversions=kversions)
        if dstdir:
            b.createrepo()
    else:
        b.build_userspace()
        b.build_kmods(kversions=kversions)
        if dstdir:
            b.createrepo()
    b.summary()



# Copyright (c) 2014-2015 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Helper to start and stop afs services."""

import os
import logging

from afsutil.system import sh, is_afs_mounted, afs_umount, unload_module, get_running, is_running

logger = logging.getLogger(__name__)

COMPONENTS = ['client', 'server']

def check_component_names(components):
    """Raises a value error if an unknown component name is given."""
    ALL = set(COMPONENTS)
    if components is None:
        c = ALL
    else:
        c = set(components)        # remove duplicates
    unknown = c.difference(ALL)    # find unknowns
    if unknown:
        s = 's' if len(unknown) > 1 else ''
        unknown = ', '.join(unknown)
        raise ValueError("Unknown component%s: %s" % (s, unknown))
    return list(c)

def _rc(component, action):
    # Try systemd if a unit file is present, otherwise fallback
    # to our init script.
    name = 'openafs-%s' % (component)
    unit_file = '/usr/lib/systemd/system/%s.service' % (name)
    init_script = "/etc/init.d/%s" % (name)
    if os.path.isfile(unit_file):
        sh('systemctl', action, name, output=False)
    elif os.path.isfile(init_script):
        if logger.getEffectiveLevel() == logging.DEBUG:
            sh('/bin/bash', '-x', init_script, action, output=False)
        else:
            sh(init_script, action, output=False)
    else:
        raise AssertionError("Init script is missing for %s!" % (name))

def start(**kwargs):
    components = check_component_names(kwargs['components'])
    if 'server' in components:
        if not is_running('bosserver'):
            _rc('server', 'start')
        else:
            logger.info("bosserver already running")
    if 'client' in components:
        if not is_afs_mounted():
            _rc('client', 'start')
        else:
            logger.info("afs already mounted")
        if not is_afs_mounted():
            raise AssertionError("Failed to start.")

def stop(**kwargs):
    components = check_component_names(kwargs['components'])
    if 'client' in components:
        if is_afs_mounted():
            _rc('client', 'stop')
        else:
            logger.info("afs already unmounted")
        if is_afs_mounted():
            afs_umount() # Be sure afs is unmounted before trying to unload.
            unload_module()
        if is_afs_mounted():
            raise AssertionError("Failed to stop.")
    if 'server' in components:
        if is_running('bosserver'):
            _rc('server', 'stop')
        servers = ('bosserver', 'upserver', 'upclient',
                   'buserver', 'bucoord', 'butc',
                   'vlserver', 'ptserver',
                   'fileserver', 'volserver',
                   'dafileserver', 'davolserver', 'salvageserver')
        still_running = get_running().intersection(servers)
        if still_running:
            raise AssertionError("Servers still running! %s" % (" ".join(list(still_running))))




# Copyright (c) 2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Create command line subcommands with argparse

This module provides a thin wrapper over the standard argparse package to
declaratively create command line subcommands. A decorator turns regular
functions into cli subcommands. The argument function defines command line
options.

Example:

    @subcommand(
        argument("host", metavar="<host>", help="example positional option"),
        argument("--filename", "-f", help="example optional flag"),
        argument("--output", default="example", help="example option"),
        )
    def example(**args):
        'example subcommand'
        print("args:", args['host'], args['filename'], args['output'])
        return 0

    dispatch()
    # usage: cli example [options]

"""


from __future__ import print_function
import argparse, logging, os, sys
from afsutil.system import CommandFailed
try:
    from configparser import ConfigParser # python3
except ImportError:
    from ConfigParser import ConfigParser # python2

progname = 'afsutil'
root = argparse.ArgumentParser()
parent = root.add_subparsers(dest='subcommand')
config = ConfigParser()
config.read([
    '/etc/{0}.cfg'.format(progname),
    os.path.expanduser('~/.{0}.cfg'.format(progname)),
    '.git/{0}.cfg'.format(progname),
])

def _get_config(section, option, default, as_boolean):
    """Get the config option"""
    value = default
    if isinstance(value, list) or isinstance(value, tuple):
        subsection = "{0}.{1}".format(section, option)
        if config.has_section(subsection):
            value = []
            for k,v in config.items(subsection):
                value.append("{0}={1}".format(k,v))
        elif config.has_option(section, option):
            value = config.get(section, option).split()
    else:
        if config.has_option(section, option):
            value = config.get(section, option)
    if as_boolean and not isinstance(value, bool):
        if value is None:
            value = False
        elif value.lower() in ('yes', 'true', 'on', '1'):
            value = True
        else:
            value = False
    return value

def _long_flag(name_or_flags):
    """Find the long flag name, if one."""
    for opt in name_or_flags:
        if opt.startswith('--'):
            return opt.lstrip('--')
    return None

def argument(*name_or_flags, **options):
    """Helper to declare subcommand arguments."""
    return (name_or_flags, options)

def subcommand(*arguments, **kwargs):
    """Decorator to declare command line subcommands."""
    def decorator(function):
        name = function.__name__.strip('_')
        desc = function.__doc__
        parser = parent.add_parser(name, description=desc)
        args = list(arguments) # Local var to add common args.
        if name not in ('help', 'version'):
            args.insert(0, argument('-q', '--quiet', help='print less messages', action='store_true'))
            args.insert(1, argument('-v', '--verbose', help='print more messages', action='store_true'))
            args.insert(2, argument('-l', '--log', help='log file location'))
        for arg in args:
            name_or_flags,options = arg
            default = options.get('default') # may be a list
            flag = _long_flag(name_or_flags)
            as_boolean = options.get('action', '') in ('store_true', 'store_false')
            if flag:
                default = _get_config(name, flag, default, as_boolean)
                options['default'] = default
            if default and 'help' in options:
                if isinstance(default, list):
                    default = " ".join(default)
                options['help'] += " (default: {0})".format(default)
            parser.add_argument(*name_or_flags, **options)
        parser.set_defaults(function=function, **kwargs)
        return function
    return decorator

def usage():
    """Print a summary of the subcommands."""
    print("usage: {0} <command> [options]".format(progname))
    print("")
    print("commands:")
    for name,parser in parent.choices.items():
        print("  {name:12} {desc}".format(name=name, desc=parser.description))
    return 0

def _setup_logging(verbose=False, quiet=False, log=None, **kwargs):
    options = {
        'level':logging.INFO,
        'format':'%(message)s',
    }
    if quiet:
        options['level'] = logging.ERROR
    if verbose:
        options['level'] = logging.DEBUG
    if log:
        options['filename'] = log
        options['format'] = '%(asctime)s %(levelname)s %(message)s'
        options['filemode'] = 'w' # clobber old logs
    logging.basicConfig(**options)
    return logging.getLogger('afsutil')

def dispatch():
    """Parse arguments and dispatch the subcommand."""
    args = vars(root.parse_args())
    requires_root = args.pop('requires_root', False)
    function = args.pop('function')
    if requires_root and os.geteuid() != 0:
        sys.stderr.write("afsutil: Must run as root!\n")
        sys.exit(1)
    log = _setup_logging(**args)
    cwd = None
    chdir = args.get('chdir')
    if chdir:
        log.info("Changing to directory %s", chdir)
        cwd = os.getcwd()
        os.chdir(chdir)
    try:
        code = function(**args)
    except CommandFailed as e:
        if args.get('log') or args.get('verbose'):
            log.exception(e)
        sys.stderr.write("Command failed: %s, code %d\n" % (e.cmd, e.code))
        sys.stderr.write("output:\n")
        sys.stderr.write("%s\n" % (e.out))
        code = 1
    finally:
        if cwd:
            os.chdir(cwd)
    return code



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Setup a new OpenAFS cell

Configure a new OpenAFS cell on one or more hosts and add fileservers to an
existing OpenAFS cell.  The OpenAFS clients and servers programs must already
be installed on the hosts. The Kerberos service key must be set and the
bosserver should be running on each server host. An OpenAFS client is required
to mount the volumes and set the ACLs of the top level volumes. The client may
be running on a separate host than the server.
"""

import time
import logging
import os
import re
import socket

import afsutil.system
import afsutil.keytab
from afsutil.cmd import bos, vos, pts, fs, udebug, rxdebug
from afsutil.system import CommandFailed, afs_mountpoint
from afsutil.transarc import AFS_SRV_LIBEXEC_DIR
from afsutil.misc import lists2dict, uniq

logger = logging.getLogger(__name__)

DBNAMES = ('ptserver', 'vlserver')
PORT = {
    'fileserver': '7000',
    'ptserver':   '7002',
    'vlserver':   '7003',
    'volserver':  '7005',
    'bosserver':  '7007',
}

class Host(object):
    """Helper to configure an OpenAFS server using the bos command."""

    def __init__(self, hostname='localhost', options=None, **kwargs):
        """Initialize the afs host object.

        hostname: hostname
        options:  dict of server options
        """
        if hostname is None or hostname == 'localhost':
            hostname = socket.gethostname()
        hostname = hostname.strip()
        self.hostname = hostname
        # The following are retrieved with bos as needed.
        self.cellname = None
        self.cellhosts = None
        self.users = None
        # Server program options.
        if options is None:
            options = {}
        self.options = options
        # Determine if this host is using dafs based on the given server
        # options. First check the host specific options, then check the host
        # independent options, finally, default to dafs.
        self.dafs = None
        if self.dafs is None:
            for program in ('dafileserver', 'davolserver', 'salvageserver', 'dasalvager'):
                flags = options.get(hostname + ":" + program, None)
                if not flags is None:
                    self.dafs = True
                    break
        if self.dafs is None:
            for program in ('fileserver', 'volserver' 'salvager'):
                flags = options.get(hostname + ":" + program, None)
                if not flags is None:
                    self.dafs = False
                    break
        if self.dafs is None:
            for program in ('dafileserver', 'davolserver', 'salvageserver', 'dasalvager'):
                flags = options.get(program, None)
                if not flags is None:
                    self.dafs = True
                    break
        if self.dafs is None:
            for program in ('fileserver', 'volserver' 'salvager'):
                flags = options.get(program, None)
                if not flags is None:
                    self.dafs = False
                    break
        if self.dafs is None:
            self.dafs = True

    def cmd(self, program):
        """Get the bos create -cmd argument."""
        # Use the canonical path; bosserver will convert it to the actual path.
        cmd = os.path.join(AFS_SRV_LIBEXEC_DIR, program)
        flags = self.options.get(self.hostname + "." + program, None)
        if flags is None:
            flags = self.options.get(program, None)
        if flags:
            cmd += ' ' + flags
        return cmd

    def rxping(self, service='bosserver', retry=10):
        try:
            rxdebug('-server', self.hostname, '-port', PORT[service], '-version', retry=retry)
        except CommandFailed:
            success = False
        else:
            success = True
        return success

    def cellinfo(self):
        """Retrieve the cell info."""
        if self.cellname is None or self.cellhosts is None:
            output = bos('listhosts', '-server', self.hostname)
            self.cellhosts = set()
            for line in output.splitlines():
                match = re.match(r'Cell name is (\S+)', line)
                if match:
                    self.cellname = match.group(1)
                match = re.match(r'\s+Host \S+ is (\S+)', line)
                if match:
                    self.cellhosts.add(match.group(1))
        return (self.cellname, tuple(self.cellhosts))

    def services(self):
        """Retrieve service names and current status."""
        output = bos('status', '-server', self.hostname, '-long')
        services = {}
        bnode = None
        bstatus = None
        for line in output.splitlines():
            match = re.match(r'Instance ([^,]+),', line)
            if match:
                bnode = match.group(1)
                match = re.search(r'currently (\w+)', line)
                if match:
                    bstatus = match.group(1)
                else:
                    bstatus = 'unknown'
                services[bnode] = {'status':bstatus}
                bnode = None
                bstatus = None
        return services

    def getservice(self, name):
        """Return the status information by name."""
        return self.services().get(name, None)

    def wait_for_status(self, name, target='running', attempts=30, delay=5):
        """Wait for service to reach running state."""
        status = 'unknown'
        logger.info("Waiting for service %s to reach %s on host %s.", name, target, self.hostname)
        for attempt in xrange(0, attempts+1):
            service = self.getservice(name)
            if service is None:
                time.sleep(delay)
                continue
            if service['status'] == target:
                logger.info("Service %s is running on host %s.", name, self.hostname)
                return
            time.sleep(delay)
        if status is None:
            status = 'unknown'
        raise AssertionError("Service %s failed to start on %s; status=%s" % (name, self.hostname, status))

    def getcellname(self):
        """Get the configured cell name for this host (ThisCell)."""
        self.cellinfo()
        return self.cellname

    def getcellhosts(self):
        """Get the configured cell hosts for this host (CellServDB)."""
        self.cellinfo()
        return tuple(self.cellhosts)

    def setcellname(self, name):
        """Set the configured cell name for this host (ThisCell)."""
        bos('setcellname', '-server', self.hostname, '-name', name)
        if self.getcellname() != name:
            raise AssertionError("Failed to update cell name!")

    def setcellhosts(self, hosts):
        """Set the configured cell hosts for this host (CellServDB).

        hosts : a sequence of hostname strings or Host objects
        """
        assert hosts is not None
        assert not isinstance(hosts, basestring) # expect a list or tuple

        hostnames = set()
        for h in hosts:
            if isinstance(h, Host):
                hostnames.add(h.hostname)
            elif isinstance(h, basestring):
                hostnames.add(h)
            else:
                raise AssertionError("Expected a Host object or string.")

        self.cellinfo()
        newhosts = hostnames - self.cellhosts
        oldhosts = self.cellhosts - hostnames
        if newhosts or oldhosts:
            logging.info("Setting cell hosts to %s", ",".join(hostnames))
            for newhost in newhosts:
                bos('addhost', '-server', self.hostname, '-host', newhost)
            for oldhost in oldhosts:
                bos('removehost', '-server', self.hostname, '-host', oldhost)
            self.cellname = None
            self.cellhosts = None
            self.cellinfo()
            if self.cellhosts != hostnames:
                raise AssertionError("Failed to update cell hosts!")

    def listusers(self):
        output = bos('listusers', '-server', self.hostname)
        return output.replace('SUsers are:', '').split()

    def adduser(self, name):
        users = self.listusers()
        if name not in users:
            logger.info("Adding %s to the superuser list on %s.", name, self.hostname)
            bos('adduser', '-server', self.hostname, '-user', name)

    def create_database(self, name):
        """Start the database server."""
        service = self.getservice(name)
        if service is not None:
            status = service['status']
            if status == 'running':
                # XXX: Update options if needed!
                logger.info("Skipping create database for %s; already running.", name)
                return
            if status == 'shutdown':
                logger.info("Retarting database for %s.", name)
                self.restart(name)
                return
            raise AssertionError("Unexpected status '%s' while trying to create db service %s on host %s." % \
                                (status, name, self.hostname))
        bos('create', '-server', self.hostname,
            '-instance', name, '-type', 'simple',
            '-cmd', self.cmd(name))

    def shutdown(self, name):
        """Shutdown the service name."""
        bos('shutdown', '-server', self.hostname, '-instance', name, '-wait')

    def shutdown_all(self):
        """Shutdown all running services."""
        for name,service in self.services().items():
            if service['status'] == 'running':
                self.shutdown(name)

    def restart(self, name):
        bos('restart', '-server', self.hostname, '-instance', name)

    def is_recovered_sync_site(self, name):
        """Returns true if this is the db sync site and the recovery state is good."""
        port = PORT[name]
        try:
            output = udebug('-server', self.hostname, '-port', port, retry=10)
        except CommandFailed:
            return False
        logger.debug("udebug for %s: %s", self.hostname, output)
        if re.search(r'clock may be bad', output):
            logger.info("Clock may be bad on host %s.", self.hostname)
        if re.search(r'I am not sync site', output):
            return False
        if re.search(r'I am sync site', output):
             match = re.search(r'Recovery state (\S+)', output)
             if match:
                 recovery_state = match.group(1)
             else:
                 recovery_state = '??'
             if recovery_state == '1f' or recovery_state == 'f':
                 logger.info("Database quorum reached for %s; sync site is %s; recovery state is %s.",
                             name, self.hostname, recovery_state)
                 return True
             logger.debug("Host %s is sync site with recovery state %s", self.hostname, recovery_state)
        return False

    def create_fileserver(self):
        """Start the fileserver (fs or dafs)."""
        bnode = 'dafs' if self.dafs else 'fs'
        service = self.getservice(bnode)
        if service is not None:
            status = service['status']
            if status == 'running':
                # XXX: Update options if needed!
                logger.info("Skipping create fileserver; already running.")
                return
            if status == 'shutdown':
                logger.info("Retarting fileserver.")
                self.restart(bnode)
                return
            raise AssertionError("Unexpected status '%s' while trying to create fileserver on host %s." % \
                                (status, self.hostname))
        if self.dafs:
            bos('create', '-server', self.hostname,
                '-instance', 'dafs', '-type', 'dafs',
                '-cmd', self.cmd('dafileserver'),
                '-cmd', self.cmd('davolserver'),
                '-cmd', self.cmd('salvageserver'),
                '-cmd', self.cmd('dasalvager'))
        else:
            bos('create', '-server', self.hostname,
                '-instance', 'fs', '-type', 'fs',
                '-cmd', self.cmd('fileserver'),
                '-cmd', self.cmd('volserver'),
                '-cmd', self.cmd('salvager'))

        ok = self.rxping(service='fileserver', retry=60)
        if not ok:
            raise AssertionError("Unable to contact file server at %s." % (self.hostname))
        ok = self.rxping(service='volserver', retry=60)
        if not ok:
            raise AssertionError("Unable to contact volume server at %s." % (self.hostname))

    def _check_volume(self, name):
        try:
            vos('listvldb', '-name', name, '-quiet', '-noresolve', '-nosort')
        except CommandFailed:
            return False
        else:
            return True

    def create_volume(self, name, partition="a"):
        """Create volume if it does not exist."""
        if self._check_volume(name):
            logger.info("Skipping create volume '%s'; already exists.", name)
        else:
            logger.info("Creating volume %s on host %s, partition %s.", name, self.hostname, partition)
            vos('create', '-server', self.hostname, '-partition', partition, '-name', name, retry=60, wait=10)

class Cell(object):

    def __init__(self, cell='localcell', db=None, fs=None,
                 admins=None, admin='admin',
                 options=None, paths=None,
                 keytab='/tmp/afs.keytab', realm=None, akimpersonate=False,
                 **kwargs):
        """Initialize the cell object.

        The first list element of the db list will be the first db server
        created. The first element of the fs list will be the first fileserver,
        which will house the rw root volumes.
        """
        # Some sanity checking.
        assert cell is not None and isinstance(cell, basestring)  # expect a string
        assert db is None or not isinstance(db, basestring) # expect a list or tuple
        assert fs is None or not isinstance(fs, basestring) # expect a list or tuple
        assert admins is None or not isinstance(admins, basestring) # expect a list or tuple
        assert options is None or not isinstance(options, basestring) # expect a list or dict
        assert paths is None or not isinstance(options, basestring) # expect a list or dict

        # Defaults.
        hostname = socket.gethostname()
        if admins is None:
            admins = [admin]
        if db is None or len(db) == 0:
            db = [hostname]
        if fs is None or len(fs) == 0:
            fs = [hostname]
        # In case 'localhost' is given as a hostname, convert to the real name.
        db = [hostname if x=='localhost' else x for x in db]
        fs = [hostname if x=='localhost' else x for x in fs]

        # Cell info.
        self.cell = cell
        self.keytab = keytab
        if realm is None:
            self.realm = cell.upper()
        else:
            self.realm = realm
        self.options = lists2dict(options)
        self.paths = lists2dict(paths)

        # Super users for this cell. Convert k5 style names to k4 style for AFS.
        self.admins = [name.replace('/', '.') for name in admins]

        # Create the set of hosts objects for this cell. A given host may
        # be a db server, a file server, or both. Avoid creating duplicate
        # objects.  The first element in the given lists are the first
        # servers for the cell setup.
        hosts = {} # tmp dict to setup db and fs lists.
        for name in uniq(db + fs):
            hosts[name] = Host(name, options=self.options)
        self.hosts = hosts.values() # list of Host objects for db and/or fs
        self.db = [hosts[name] for name in uniq(db)]
        self.fs = [hosts[name] for name in uniq(fs)]

        # Set the programs to be used by login().
        self.akimpersonate = akimpersonate
        for cmd in self.paths:
            afsutil.cmd.setpath(cmd, self.paths[cmd])

    @classmethod
    def current(cls, **kwargs):
        """Create a cell object from the existing cell."""
        host = Host()
        cell = kwargs.pop('cell', host.getcellname())
        db = kwargs.pop('db', host.getcellhosts())
        fs = kwargs.pop('fs', ['localhost']) # get list from vos listaddrs?
        admins = kwargs.pop('admins', host.listusers())
        return cls(cell=cell, db=db, fs=fs, admins=admins, **kwargs)

    def _akimpersonate(self, user):
        if not os.path.exists(self.keytab):
            raise ValueError("Keytab file not found.")
        k = afsutil.keytab.Keytab.load(self.keytab)
        k.akimpersonate(user=user, cell=self.cell, realm=self.realm)

    def _kinit_aklog(self, user):
        if not os.path.exists(self.keytab):
            raise ValueError("Keytab file not found.")
        output = afsutil.cmd.kinit('-V', '-k', '-t', self.keytab, user.replace('.','/'))
        for line in output.splitlines():
            logger.info(line)
        output = afsutil.cmd.aklog('-d', '-c', self.cell, '-k', self.realm)
        for line in output.splitlines():
            logger.info(line)

    def _wait_for_quorum(self, name, hosts, attempts=60, delay=10):
        for attempt in xrange(0, attempts+1):
            logger.info("Waiting for %s database quorum; attempt %d of %d.", name, (attempt+1), attempts)
            num_sync_sites = 0
            for host in hosts:
                if host.is_recovered_sync_site(name):
                    num_sync_sites += 1
            if num_sync_sites == 1:
                logger.debug("Database quorum reached for %s.", name)
                return
            time.sleep(delay)
        raise AssertionError("Failed to reach database quorum for %s." % (name))

    def _create_admin(self, admin):
        logger.info("Creating the admin user %s.", admin)
        # Due to a bug in some versions of OpenAFS, the db drops quorum the
        # first time we write to it after the first election. Wait at least
        # ubik BIGTIME and retry.
        try:
            pts('createuser', '-name', admin, retry=1, wait=80)
        except CommandFailed as e:
            if not "Entry for name already exists" in e.out:
                raise
        try:
            pts('adduser', '-user', admin, '-group', 'system:administrators')
        except CommandFailed as e:
            if not "Entry for id already exists" in e.out:
                raise

    def _wscell(self):
        output = fs('wscell')  # Raises exception if client is not running.
        match = re.match("This workstation belongs to cell '(\S+)'", output)
        if match:
            cell = match.group(1)
        else:
            raise AssertionError("Unable to get local cell name!")
        logger.info("Cell is %s", self.cell)
        return cell

    def _mount(self, path, volume, *opts):
        if not os.path.exists(path):
            msg = ' as read-write' if '-rw' in opts else ''
            logger.info("Mounting '%s' on path '%s'%s.", volume, path, msg)
            fs('mkmount', '-dir', path, '-vol', volume, *opts)

    def _create_replica(self, name):
        output = vos('listvldb', '-name', name, '-quiet')
        if re.findall(r'^\s+server \S+ partition \S+ RO Site', output, re.M):
            logger.info("Skipping replication of %s; already have a read only site", name)
            return
        server,partition = re.findall(r'^\s+server (\S+) partition (\S+) RW Site', output, re.M)[0]
        # Due to a bug in some versions of OpenAFS, the db drops quorum the
        # first time we write to it after the first election. vos fails with a
        # uquorum error and and the volume is left locked. Use this closure to
        # unlock the the volume before retrying the vos command.
        def _unlocker(name):
            def _unlock():
                try:
                    vos('unlock', '-id', name)
                except:
                    pass
            return _unlock
        vos('addsite', '-server', server, '-partition', partition, '-id', name,
            retry=20, wait=80, cleanup=_unlocker(name))
        vos('release', '-id', name,
            retry=20, wait=80, cleanup=_unlocker(name))

    def ping_hosts(self):
        """Verify hosts are reachable and bosserver is running."""
        failed = []
        for host in self.hosts:
            ok = host.rxping(service='bosserver', retry=0)
            if not ok:
                failed.append(host.hostname)
        if failed:
            s = 's' if len(failed) > 1 else ''
            f = ', '.join(failed)
            raise AssertionError("Failed to reach bosserver on host%s %s" % (s, f))

    def _shutdown_hosts(self):
        """Shutdown the services on all the hosts."""
        for host in self.hosts:
            host.shutdown_all()

    def _setup_first_db_server(self):
        """Setup the initial database server and db files."""
        db = self.db[0]
        logger.info("Setting up the first database server %s." % db.hostname)

        # Setup the cell info for a single database server so the empty db can
        # be created and quorum established.
        db.setcellname(self.cell)
        db.setcellhosts([self.db[0]])
        for dbname in DBNAMES:
            db.create_database(dbname)
            db.wait_for_status(dbname, target='running')

        logger.info("Waiting for quorum.")
        time.sleep(2) # Give the servers a chance to start
        for dbname in DBNAMES:
            self._wait_for_quorum(dbname, [db])

        # The database servers create emtpy prdb and vldb databases as
        # side-effect of these queries, including the creation of the initial
        # ubik database versions.
        pts('listentries', retry=10)
        vos('listvldb', retry=10)

        # Create the superusers and add them to this first server's userlist.
        for admin in self.admins:
            self._create_admin(admin)
            db.adduser(admin)

    def _add_db_servers(self):
        """Setup the remaining database servers."""
        logger.info("Setting up additional database servers.")

        # Shutdown the primary server since we are changing the
        # cell hosts on it.
        for dbname in DBNAMES:
            self.db[0].shutdown(dbname)
            self.db[0].wait_for_status(dbname, target='shutdown')
        time.sleep(1)

        # Set the cell hosts on all the db servers, including the primary.
        # Be sure keep the order of the hosts consistent, since that is
        # the normal setup.
        for host in self.db:
            host.setcellname(self.cell)
            host.setcellhosts([self.db[0]])
            host.setcellhosts(self.db)

        # Restart the primary and create the other database hosts.
        # Use udebug to verify quorum is established.
        for dbname in DBNAMES:
            self.db[0].restart(dbname)
            self.db[0].wait_for_status(dbname, target='running')
            for host in self.db:
                if host != self.db[0]:
                    for admin in self.admins:
                        host.adduser(admin)
                    host.create_database(dbname)
                    host.wait_for_status(dbname, target='running')

        logger.info("Waiting for quorum.")
        time.sleep(15)
        for dbname in DBNAMES:
            self._wait_for_quorum(dbname, self.db)

    def _add_fs_servers(self):
        """Add remaining file servers."""
        for server in self.fs:
            if server != self.fs[0]:
                self.addfs(server)

    def _setup_first_fs_server(self):
        """Startup the file server processes and create the root volumes if needed."""
        fs = self.fs[0]
        logger.info("Setting up the first file server %s." % fs.hostname)
        if fs != self.db[0]:
            fs.setcellname(self.cell)
            fs.setcellhosts([self.db[0]])
            for admin in self.admins:
                fs.adduser(admin)
        fs.create_fileserver()
        bnode = 'dafs' if fs.dafs else 'fs'
        fs.wait_for_status(bnode, target='running')

        # Note: root.afs must exist before non-dynroot clients are started.
        fs.create_volume('root.afs')
        fs.create_volume('root.cell')

    def login(self, user):
        """Obtain a token for this cell.

        This function may be invoked after newcell() on a running client
        to obtain an AFS token.
        """
        if self.akimpersonate:
            self._akimpersonate(user)
        else:
            self._kinit_aklog(user)

    def newcell(self):
        """Setup a new OpenAFS cell.

        This is the server side setup of the new cell.  This function should be
        run on a new system which has the servers installed, and the bosserver
        running. The OpenAFS commands are run with the -localauth option to
        configure the cell. Only OpenAFS commands are used to access remote
        systems, so any remote configuration which requires remote shell access
        (e.g. ssh) must be done prior to running newcell().

        Preconditions:
        1. The vicep partitions must exist and be empty on each host.
        2. The OpenAFS server programs must be installed on the localhost.
        3. The OpenAFS server programs must be installed on the remote hosts.
        4. The NetInfo/NetRestict configuration must be completed on each host.
        5. The Kerberos service key must be set on each host.
        6. The OpenAFS bosserver must be running on each host.
        7. The OpenAFS client must be installed on the localhost.
        8. The OpenAFS client cell config must be set (ThisCell and CellServDB)
        9. The OpenAFS client must be already running in dynroot mode *OR*
           must be startable with afsutil.system.start()

        Postconditions:
        1. The database servers are configured and running.
        2. The fileservers are configured and running.
        3. A set of admin and regular users are created.
        """
        logger.info("Setting up new cell.")
        self.ping_hosts()      # bosserver must be running on each
        self._shutdown_hosts() # before the CellServDBs are changed
        self._setup_first_db_server()
        self._setup_first_fs_server()
        if len(self.db) > 1:
            self._add_db_servers()
        if len(self.fs) > 1:
            self._add_fs_servers()

    def addfs(self, host):
        """Add a fileserver to this cell.

        The remote host must have the binaries installed, the service key
        installed, the bosserver started.  This function will setup the
        cell name, the CellServDB configuration, add the superuser names,
        and then create the bosserver configuration to run the fileserver.
        """
        if isinstance(host, basestring):
            host = Host(host, options=self.options)
        logger.info("Adding fileserver %s", host.hostname)
        host.setcellname(self.cell)
        host.setcellhosts(self.db)
        for admin in self.admins:
            host.adduser(admin)
        host.create_fileserver()

    def mtroot(self, volumes):
        """Mount, setup acls, and replicate the root.afs and root.cell volumes.

        This function should be run on a system which has a running cache
        manager, after the Cell.newcell() function has been invoked on the
        server.

        Preconditions:
        1. The root.afs and root.cell volumes must exist.
        2. The servers must be running (on the local or remote hosts).
        3. The cache manager must be running (with or without -dynroot).

        Post-conditions:
        1. The root.afs.readonly and root.cell.readonly volumes exist.
        2. The path /afs/<cellname> resolves to the root.cell.readonly root vnode.
        3. The path /afs/.<cellname> resolves to the root.cell root vnode.
        4. The path /afs/.<cellname>/.afs resolves to the root.afs root vnode.
        5. The optional list of given volumes are mounted and replicated under
           /afs/<cellname>

        The special mount point '/afs/.<cellname>/.afs' is created in order to
        to add mount points to the root.cell volume on cache managers running
        dynroot mode.

        Note: In the past, this function attempted to create the cellular mount
        point using the magic '.:mount' path when the cache manager was running
        in dynroot mode.  However we found this did not work on Solaris unless
        the -fakestat option was also given. This is arguably an OpenAFS bug.

        Note: It would be nice if dynroot mode could be could be detected, but
        the OpenAFS cache manager does not provide a nice and portable way to
        do so at this time, so we pass the afsd options used during the cache
        manager install to detect if dynroot is enabled.
        """
        # First, verify AFS is mounted and find the mount point to it, e.g., /afs
        afs = afs_mountpoint()
        if afs is None:
            raise AssertionError("Unable to mount volumes; afs is not mounted!")

        # Verify the cache manager's cell matches the name of the cell
        # being setup. This mismatch can happen if the cache manager's ThisCell
        # and/or CellServDB files were not created correctly before the cache
        # was started.
        cell = self._wscell()
        if cell != self.cell:
            raise AssertionError("Client side ThisCell file does not match the cell name!")

        # Get an admin token to mount the volumes and set the acls.
        user = self.admins[0]
        self.login(user)

        # Replicate our root volumes.
        self._create_replica('root.afs')
        self._create_replica('root.cell')

        # Mount the root volumes.
        afsd_options = self.options.get('afsd', '')
        dynroot = '-dynroot' in afsd_options
        if not dynroot:
            self._mount("%(afs)s/%(cell)s" % locals(), 'root.cell', '-cell', cell)
            self._mount("%(afs)s/.%(cell)s" % locals(), 'root.cell', '-cell', cell, '-rw')
            self._mount("%(afs)s/.%(cell)s/.afs" % locals(), 'root.afs', '-rw') # for dynroot clients
        else:
            self._mount("%(afs)s/.%(cell)s/.afs" % locals(), 'root.afs', '-rw')
            self._mount("%(afs)s/.%(cell)s/.afs/%(cell)s" % locals(), 'root.cell', '-cell', cell)
            self._mount("%(afs)s/.%(cell)s/.afs/.%(cell)s" % locals(), 'root.cell', '-cell', cell, '-rw')
        fs('checkvolumes')

        # Grant global read and list rights to the root /afs and /afs/<cell> paths.
        if not dynroot:
            fs('setacl', '-dir', "%(afs)s" % locals(), '-acl', 'system:anyuser', 'read')
            fs('setacl', '-dir', "%(afs)s/.%(cell)s" % locals(), '-acl', 'system:anyuser', 'read')
        else:
            fs('setacl', '-dir', "%(afs)s/.%(cell)s/.afs" % locals(), '-acl', 'system:anyuser', 'read')
            fs('setacl', '-dir', "%(afs)s/.%(cell)s" % locals(), '-acl', 'system:anyuser', 'read')

        # Place top level volumes on the same fileserver as the root volumes.
        for name in volumes:
            self.fs[0].create_volume(name)
            self._create_replica(name)
            self._mount("%(afs)s/.%(cell)s/%(name)s" % locals(), name, '-cell', cell)
            self._mount("%(afs)s/.%(cell)s/.%(name)s" % locals(), name, '-cell', cell, '-rw')
            fs('setacl', '-dir', "%(afs)s/.%(cell)s/.%(name)s" % locals(), '-acl', 'system:anyuser', 'read')
        vos('release', '-id', 'root.cell')
        fs('checkvolumes')

def newcell(**kwargs):
    cell = Cell(**kwargs)
    cell.newcell()

def mtroot(**kwargs):
    top = kwargs.pop('top', [])
    cell = Cell(**kwargs)
    cell.mtroot(top)

def addfs(**kwargs):
    hostname = kwargs.pop('hostname', socket.gethostname())
    cell = Cell(**kwargs)
    cell.addfs(hostname)

def login(**kwargs):
    if os.geteuid() == 0:
        logger.warning("Running afsutil login as root! Your regular user will not have a token.")
    user = kwargs.pop('user', 'admin')
    cell = Cell(**kwargs)
    cell.login(user)



# Copyright (c) 2014-2018 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Install build dependencies for OpenAFS."""

import logging
import os
import platform
import urllib2
import tempfile

from afsutil.system import sh, CommandFailed

logger = logging.getLogger(__name__)

class Unsupported(Exception):
    pass

def apt_get_install(packages, dryrun=False):
    sh('apt-get', '-y', 'install', *packages, dryrun=dryrun, output=False)

def yum_install(packages, dryrun=False):
    sh('yum', 'install', '-y', *packages, dryrun=dryrun, output=False)

def dnf_install(packages, dryrun=False):
    sh('dnf', 'install', '-y', *packages, dryrun=dryrun, output=False)

def zypper_install(packages, dryrun=False):
    sh('zypper', 'install', '-y', *packages, dryrun=dryrun, output=False)

def pkg_install(packages, dryrun=False, update_all=True):
    # Recent versions of solarisstudio fail to install due to dependency
    # conflicts unless the system is updated first.
    if update_all:
        sh('pkg', 'update', '-v', 'entire@latest', dryrun=dryrun, output=False)
    # Ignore package already installed errors.
    ERROR_ALREADY_INSTALLED = 4
    try:
        sh('pkg', 'install', '--accept', *packages, dryrun=dryrun, output=False)
    except CommandFailed as e:
        if e.code != ERROR_ALREADY_INSTALLED:
            logger.error("pkg install failed: %s" % e)
            raise e

def lookup_solarisstudio(creds='/root/creds',
                         key='pkg.oracle.com.key.pem',
                         cert='pkg.oracle.com.certificate.pem',
                         **kwargs):
    """
    Lookup the Solaris Studio package.

    Before running this function, create an account on the Oracle Technology Network
    and follow the instructions to create and download the key and certificate files.
    Place the key and cert file in a local directory or at a private url.  The
    default location of the key and cert files is /root/creds. Set the 'creds'
    keyword argurment to specify a different path or a url.
    """
    path = None
    tmpdir = None
    quiet = True

    def download(baseurl, filename, path):
        url = os.path.join(baseurl, filename)
        dst = os.path.join(path, filename)
        rsp = urllib2.urlopen(url)
        logger.info("Downloading '%s' to '%s'.", url, dst)
        with open(dst, 'wb') as f:
            f.write(rsp.read())

    def normalize(s):
        return [int(x) for x in s.split('.')]

    def compare_versions(a, b):
        return cmp(normalize(a), normalize(b))

    if creds.startswith('http://') or creds.startswith('https://'):
        try:
            path = tmpdir = tempfile.mkdtemp()
            download(creds, key, tmpdir)
            download(creds, cert, tmpdir)
        except urllib2.HTTPError as e:
            logger.error("Unable to download files from url '%s', %s'." % (creds, e))
            return None
        finally:
            logger.info("Cleaning up temporary files in '%s'.", tmpdir)
            if os.path.exists(os.path.join(tmpdir, key)):
                os.remove(os.path.join(tmpdir, key))
            if os.path.exists(os.path.join(tmpdir, cert)):
                os.remove(os.path.join(tmpdir, cert))
            if os.path.exists(tmpdir) and tmpdir != "/":
                os.rmdir(tmpdir)
    elif os.path.exists(creds):
        logger.info("Using credentials in path '%s'.", creds)
        path = creds
    else:
        logger.error("creds path '%s' not found.", creds)
        return None

    logger.info("Setting publisher for solarisstudio.")
    sh('pkg', 'set-publisher',   # Will refresh if already set.
       '-k', os.path.join(path, key),
       '-c', os.path.join(path, cert),
       '-G', '*',
       '-g', 'https://pkg.oracle.com/solarisstudio/release', 'solarisstudio',
       output=False, quiet=quiet)

    logger.info("Getting available solaris studio packages.")
    packages = {}
    output = sh('pkg', 'list', '-H', '-a', '-v', '--no-refresh',
                'pkg://solarisstudio/*', quiet=quiet)
    for line in output:
        # Extract the root package name, version, and install state.
        fmri,ifo = line.split()
        pkg,vt = fmri.split('@')
        v,t = vt.split(':')
        version = v.split(',')[0]
        name = pkg.replace('pkg://solarisstudio/developer/','')
        istate = ifo[0] == 'i'
        if '/' in name:
            pass # Skip non-root ones, e.g. 'developerstudio-125/cc'.
        else:
            logger.info("Found package %s, version %s, installed %s.",
                        name, version, istate)
            packages[name] = {'version':version, 'installed':istate}

    # Find the most recent version.
    solarisstudio = None
    vers = '0.0'
    for name in packages.keys():
        if compare_versions(packages[name]['version'], vers) > 0:
            vers = packages[name]['version']
            solarisstudio = 'pkg://solarisstudio/developer/{0}'.format(name)
    if solarisstudio:
        logger.info("Most recent version found is '%s'.", solarisstudio)
    else:
        logger.info("Could not find solaris studio package.")
    return solarisstudio

def getdeps(dryrun=False, skip_headers=False, skip_solarisstudio=False, **kwargs):
    """Determine platform and install build dependencies."""
    system = platform.system().lower()
    if system == 'linux':
        dist = platform.dist()[0].lower()
        release = platform.dist()[1].lower()
        if dist == 'debian' or dist == 'ubuntu':
            packages = [
                'autoconf',
                'automake',
                'bison',
                'comerr-dev',
                'dblatex',
                'dkms',
                'docbook-xsl',
                'doxygen',
                'flex',
                'libelf-dev',
                'libfuse-dev',
                'libkrb5-dev',
                'libncurses5-dev',
                'libpam0g-dev',
                'libtool',
                'libxml2-utils',
                'perl',
                'pkg-config',
                'xsltproc',
            ]
            if not skip_headers:
                packages.append('linux-headers-{0}'.format(platform.release()))
            apt_get_install(packages, dryrun)
        elif dist == 'centos':
            packages = [
                'autoconf',
                'automake',
                'bison',
                'flex',
                'fuse-devel',
                'gcc',
                'glibc-devel',
                'krb5-devel',
                'libtool',
                'make',
                'ncurses-devel',
                'pam-devel',
                'perl-devel',
                'perl-ExtUtils-Embed',
                'perl-Test-Simple',
                'redhat-rpm-config',
                'rpm-build',
                'swig',
                'wget',
            ]
            if not skip_headers:
                packages.append('kernel-devel-{0}'.format(platform.release()))
            yum_install(packages, dryrun)
        elif dist == 'suse':
            packages = [
                'autoconf',
                'automake',
                'bison',
                'flex',
                'fuse-devel',
                'gcc',
                'glibc-devel',
                'krb5-devel',
                'libtool',
                'make',
                'ncurses-devel',
                'pam-devel',
                'rpm-build',
                'wget',
            ]
            if not skip_headers:
                packages.append('kernel-devel')
            zypper_install(packages, dryrun)
        elif dist == 'fedora' and release < 22:
            packages = [
                'autoconf',
                'automake',
                'bison',
                'flex',
                'fuse-devel',
                'gcc',
                'glibc-devel',
                'krb5-devel',
                'libtool',
                'make',
                'ncurses-devel',
                'pam-devel',
                'perl-devel',
                'perl-ExtUtils-Embed',
                'perl-Test-Simple',
                'redhat-rpm-config',
                'rpm-build',
                'wget',
            ]
            if not skip_headers:
                packages.append('kernel-devel-{0}'.format(platform.release()))
            yum_install(packages, dryrun)
        elif dist == 'fedora':
            packages = [
                'autoconf',
                'automake',
                'bison',
                'flex',
                'fuse-devel',
                'gcc',
                'glibc-devel',
                'krb5-devel',
                'libtool',
                'make',
                'ncurses-devel',
                'pam-devel',
                'perl-devel',
                'perl-ExtUtils-Embed',
                'perl-Test-Simple',
                'redhat-rpm-config',
                'rpm-build',
                'wget',
            ]
            if not skip_headers:
                packages.append('kernel-devel-{0}'.format(platform.release()))
            dnf_install(packages, dryrun)
        else:
            raise Unsupported("Linux dist '{0}'.".format(dist))
    elif system == 'sunos':
        release = platform.release()
        if release == '5.11':
            packages = [
                'autoconf',
                'automake',
                'bison',
                'flex',
                'git',
                'gnu-binutils',
                'gnu-coreutils',
                'gnu-sed',
                'libtool',
                'make',
                'onbld',
                'text/locale',
            ]
            if not skip_solarisstudio:
                solarisstudio = lookup_solarisstudio(**kwargs)
                if solarisstudio:
                    packages.append(solarisstudio)
            pkg_install(packages, dryrun)
        else:
            raise Unsupported("Solaris release '{0}'".format(release))
    else:
        raise Unsupported("Operating system '{0}'.".format(system))



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Build OpenAFS from sources (developer tool)"""

import logging
import os
import re
import shlex
import platform
import glob

from afsutil.system import sh, CommandFailed, tar, mkdirp
from afsutil.misc import lists2dict, flatten

logger = logging.getLogger(__name__)

def cfopts():
    """Return the default configure options for this system."""
    options = [
        '--enable-debug',
        '--enable-debug-kernel',
        '--without-dot',
        '--enable-transarc-paths',
    ]
    sysname = os.uname()[0]
    if sysname == "Linux":
        options.append('--with-linux-kernel-packaging')
        options.append('--enable-checking')
    return " ".join(options)

def _detect_sysname():
    """Try to detect the sysname from the previous build output."""
    sysname = None
    if os.path.exists("src/config/Makefile.config"):
        with open("src/config/Makefile.config", "r") as f:
            for line in f.readlines():
                match = re.match(r'SYS_NAME\s*=\s*(\S+)', line)
                if match:
                    sysname = match.group(1)
                    break
    return sysname

def _sanity_check_dir():
    msg = "Missing '%s'; are you in the OpenAFS source top-level directory?"
    for d in ('src', 'src/afs', 'src/viced'):
        if not os.path.isdir(d):
            raise AssertionError(msg % (d))

def is_git_clean_enabled(gitdir, git='git'):
    enabled = False
    try:
        output = sh(git, '--git-dir', gitdir, 'config', '--bool', '--get', 'afsutil.clean')
        if output[0] == 'true':
            enabled = True
    except CommandFailed as e:
        if e.code != 1:
            raise e
    return enabled

def _detect_solariscc():
    search = [
        '/opt/developerstudio*/bin',
        '/opt/solarisstudio*/bin',
    ]
    for pattern in search:
        paths = sorted(glob.glob(pattern), reverse=True)
        if len(paths) != 0:
            return paths[0]
    return None

def _setenv_solaris():
    need = [
        '/usr/perl5/bin',   # for pod2man
    ]
    # Update the path to the solaris studio cc.
    if not 'SOLARISCC' in os.environ:
        ccpath = _detect_solariscc()
        if ccpath:
            need.append(ccpath)
            solariscc = os.path.join(ccpath, 'cc')
            logger.info("Setting SOLARISCC to '%s'", solariscc)
            os.environ['SOLARISCC'] = solariscc
        else:
            logger.warning("Failed to find path to solaris cc!")
    # Update the PATH.
    paths = os.getenv('PATH', '').split(':')
    for path in need:
        if not path in paths:
            logger.info("Adding '%s' to PATH." % (path))
            paths.append(path)
    os.environ['PATH'] = ':'.join(paths)

def _setenv():
    system = platform.system()
    if system == 'SunOS':
        _setenv_solaris()

def _create_tarball(tarball=None, program=None):
    sysname = _detect_sysname()
    if sysname is None:
        raise AssertionError("Cannot find sysname.")
    if tarball is None:
        tardir = 'packages'
        if not os.path.isdir(tardir):
            mkdirp(tardir)
        tarball = os.path.join(tardir, "openafs-%s.tar.gz" % (sysname))
    tar(tarball, sysname, tar=program)
    logger.info("Created tar file %s", tarball)

def _cfrm(cf, option):
    if option in cf:
        cf.remove(option)

def _cfadd(cf, option):
    if option not in cf:
        cf.append(option)


def regen(srcdir='.', force=False):
    if not force and os.path.exists('%s/configure' % srcdir):
        logger.warning("Skipping regen.sh; configure already exists")
        return 0
    sh('/bin/sh', '-c', 'cd %s && ./regen.sh' % srcdir, output=False)

def configure(options=None, srcdir='.', force=False):
    if options is None:
        options = []
    if not force and os.path.exists('config.status'):
        logger.warning("Skipping configure; config.status already exists")
        return 0
    sh('%s/configure' % srcdir, *options, output=False)

def make(jobs=1, target='all', program='make'):
    args = [program]
    if jobs > 1:
        # Requires a `make` whichsupports the -j option. If your `make` does
        # not, then specify the path to `gmake` or set jobs to 1.
        args.append('-j')
        args.append('{0}'.format(jobs))
    args.append(target)
    sh(*args, output=False)

def build(**kwargs):
    """Build the OpenAFS binaries.

    Build the transarc-path compatible bins by default, which are
    deprecated, but old habits die hard.
    """
    cf = kwargs.get('cf', cfopts())
    xcf = flatten(kwargs.get('xcf', []))
    target = kwargs.get('target', 'all')
    clean = kwargs.get('clean', False)
    jobs = kwargs.get('jobs', 1)
    srcdir = kwargs.get('srcdir', '.')
    tarball = kwargs.get('tarball', None)
    paths = kwargs.get('paths', [])
    log = kwargs.get('log', None)

    paths = lists2dict(paths)
    _make = paths.get('make', 'make')
    _git = paths.get('git', 'git')
    _tar = paths.get('tar', None) # tar or gtar, depending on OS

    cf = shlex.split(cf)  # Note: shlex handles quoting properly.
    for x in xcf:
        for y in shlex.split(x):
            _cfadd(cf, y)

    # The legacy Transarc-style distribution requires a different top level target.
    if '--enable-transarc-paths' in cf and target == 'all':
        target = 'dest'

    if not os.path.exists(srcdir):
        raise AssertionError("srcdir not found: %s" % (srcdir))

    if srcdir == '.':
        _sanity_check_dir()

    if clean:
        gitdir = os.path.abspath(os.path.join(srcdir, '.git'))
        if not os.path.isdir(gitdir):
            logger.error("Unable to run git clean; not a git directory: '%s'", gitdir)
            return 1
        if not is_git_clean_enabled(gitdir, git=_git):
            logger.warning("Unable to run git clean; not enabled in the git config.")
            logger.warning("To enable git clean before builds:")
            logger.warning("    git config --local afsutil.clean true");
            return 1
        args = [_git, '--git-dir', gitdir, '--work-tree', srcdir, 'clean', '-f', '-d', '-x', '-q']
        if log:
            args.append('--exclude')
            args.append(log)
        sh(*args, output=False)

    _setenv()
    regen(srcdir=srcdir)
    configure(options=cf, srcdir=srcdir)
    make(jobs=jobs, target=target, program=_make)

    if target == 'dest':
        _create_tarball(tarball, program=_tar)
